from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader

import logging
import xarray as xr


class MODISDataLoader(ScalarDataLoader):        
    def import_data(self, bounds):
        '''
        Reads in data from a MODIS NetCDF file. 
        Renames variable to 'SIC'
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            xr.Dataset: 
                MODIS dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'SIC'
        '''
        logging.info(f"- Opening file {self.file}")
        # Open Dataset
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files)
        # Change column name
        data = data.rename({'iceArea': 'SIC'})

        # Set areas obscured by cloud to NaN values
        data = data.where(data.cloud != 1, drop=True)
        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)
        
        return data



from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader

import xarray as xr


class BalticSeaIceDataLoader(ScalarDataLoader):
    def import_data(self, bounds):
        '''
        Reads in data from a Baltic Sea Ice NetCDF file. 
        Renames coordinates to 'lat' and 'long', and renames variable to 'SIC'
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            xr.Dataset: 
                Baltic Sea Ice dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'SIC'         
        '''
        # Open Dataset
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files)
        # Change column names
        data = data.rename({'ice_concentration': 'SIC',
                            'lon': 'long'})
        # Limit to just SIC data
        data = data['SIC'].to_dataset()
        # Reverse order of lat as array goes from max to min
        data = data.reindex(lat=data.lat[::-1])
        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)
        
        return data








class Metadata:
    """
    A Metadata is a class that defines the datasource for a certain Cellbox and the assocated splitting conditions.


    Attributes:
       data_loader : object of the DataLoader class that enable projecting to the cellbox data
       splitting_conditions: list of conditions that determine how to split the data accessed by data_loader 
       value_fill_type (string): indicates how to fill a CellBox if it has void data
         (ex. use the data in the parent cellbox or assign 0 to the data)

    """
   

    def __init__(self, data_loader , spliiting_conditions =None, value_fill_type= ""):
        """

            Args:
               data_loader (DataLoader): object of the DataLoader class that enables projecting to the cellbox data
               splitting_conditions (List<dict>): list of conditions that determine how to split CellBox
               value_fill_tyep (string): represents the way the data of a cellbox will be filled
                 in case it has void data (ex. parent , 0 )
                
        """ 
        self.data_loader = data_loader
        self.splitting_conditions = spliiting_conditions
        self.value_fill_type = value_fill_type
  
    def get_data_loader(self): 
        """
        returns the data loader
        """
        return self.data_loader

    def get_splitting_conditions(self):
        """
        returns a list of the splitting conditions
        """
        
        return self.splitting_conditions

    def set_data_loader(self , data_loader): 
        """
        sets the data loader
        """
        self.data_loader = data_loader
 

    def set_splitting_conditions(self ,  splitting_conditions):
        """
        sets the splitting conditions
        """
        self.splitting_conditions = splitting_conditions

    def set_value_fill_type(self , value_fill_type): 
        """
        sets the value fill type
        """
        self.value_fill_type = value_fill_type

    def get_value_fill_type(self ):
        """
        returns thevalue fill type
        """
        return self.value_fill_type   
  




from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader

import logging

from datetime import datetime, timedelta

import xarray as xr
from pandas import to_timedelta
from numpy import datetime64


class IceNetDataLoader(ScalarDataLoader):
    def import_data(self, bounds):
        '''
        Reads in data from a IceNet 2 NetCDF file. 
        Renames coordinates to 'lat' and 'long', and renames variable to 
        'SIC'
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            pd.DataFrame: 
                IceNet dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'SIC'
        '''
        def filename_to_datetime(filename):
            '''
            Converts icenet filenames to datetime objects
            Assumes file names are in the format 
            <hemisphere>_daily_forecast.<YYYY-MM-DD>.nc
            '''
            return datetime.strptime(filename.split('.')[1], '%Y-%m-%d')
        
        # Convert temporal boundary to datetime objects for comparison
        max_time = datetime.strptime(bounds.get_time_max(), '%Y-%m-%d')
        min_time = datetime.strptime(bounds.get_time_min(), '%Y-%m-%d')
        time_range = max_time - min_time
        # Retrieve list of dates from filenames
        file_dates = {filename_to_datetime(file): file 
                      for file in self.files 
                      if filename_to_datetime(file) < min_time}
        # Find closest date prior to min_time
        closest_date = max(k for k, v in file_dates.iteritems())

        # Open Dataset
        ds = xr.open_dataset(file_dates[closest_date])
        # Cast coordinates/variables to those understood by mesh
        ds = ds.rename({'lon':'long',
                        'sic_mean': 'SIC'})
        
        # Max number of days in future IceNet can predict
        max_leadtime = int(ds.leadtime.max())
        
        # Ensure that temporal boundary is possible before extracting
        assert time_range < timedelta(days=max_leadtime),\
            f'Time boundary too large! Forecast only runs for max of {max_leadtime} days'
        
        assert closest_date + timedelta(days=max_leadtime) > max_time,\
            'Time boundary runs beyond max forecast date!'
        
        logging.info(f"- Searching for closest date prior to {bounds.get_time_min()}")
        # For the days in forecast range of IceNet dataset
        for days_ago in range(1, max_leadtime+1):
            # Set the date from which the forecast is taken
            start_time  = datetime64(min_time - timedelta(days=days_ago))
            try:
                # See if day exists, raises error if date not in dataset
                ds = ds.sel(time=start_time)
                break
            except:
                # Error thrown, date not in dataset. Try previous day
                logging.debug(f'\tUnable to select start day of {start_time} for IceNet, trying previous day')
                continue
        else:
            # If ran through entire dataset with no valid dates
            raise EOFError('No valid start date found in IceNet data!')
        
        assert (time_range.days < max_leadtime - days_ago),\
            f'''Not enough leadtime to support date range specified!
            End ({max_time}) - Start({min_time}) = {time_range.days} days
            Leadtime ({max_leadtime}) days - Prediction({days_ago}) days ago = {max_leadtime-days_ago} days
            '''
        
        # TODO fix logging bug.
        #logging.info(f"- Found date {datetime.strftime('%Y-%m-%d')}")

        # Choose predictions from earliest date before start_date
        ds = ds.sel(leadtime=range(days_ago, time_range.days + days_ago))
        # Set to pd.DataFrame so can limit by lat/long
        df = ds.to_dataframe().reset_index()
        # Set time column to be dates of predictions
        # rather than date on which prediction made
        df.time = df.time + to_timedelta(df.leadtime, unit='d')
        # Remove unwanted columns
        df = df.drop(columns=['yc','xc','leadtime', 'Lambert_Azimuthal_Grid', 'sic_stddev', 'forecast_date','ensemble_members'])
        # Trim to initial datapoints
        df = self.trim_datapoints(bounds, data=df)
        
        # Turn SIC into a percentage
        df.SIC = df.SIC.apply(lambda x: x*100)
        
        # Return extracted data
        return df



from meshiphi.dataloaders.vector.abstract_vector import VectorDataLoader

import logging

import xarray as xr

from datetime import datetime
from os.path import basename

class DuacsCurrentDataLoader(VectorDataLoader):
    def import_data(self, bounds):
        """
        Reads in data from a DUACS altimeter derived current NetCDF file.
        Renames coordinates to 'lat' and 'long', and renames variable to
        'uC, vC'

        Args:
            bounds (Boundary): Initial boundary to limit the dataset to

        Returns:
            xr.Dataset:
                Near real-time current dataset within limits of bounds.
                Dataset has coordinates 'lat', 'long', and variable 'uC', 'vC'
        """
        time_range = [datetime.strptime(time_str, "%Y-%m-%d") 
                      for time_str in bounds.get_time_range()]
        # Reduce files to those within date range
        self.files = [file for file in self.files 
                      if time_range[0] \
                      <= datetime.strptime(basename(file)[10:-3], "%Y-%m-%d") \
                      <= time_range[1]]
        
        # Open Dataset
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files).compute()
        # Change column names
        data = data.rename({'latitude': 'lat',
                            'longitude': 'long',
                            'ugos': 'uC',
                            'vgos': 'vC'})
        # Drop unnecessary variable
        data = data.drop_vars('crs')

        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)

        return data





from meshiphi.mesh_generation.direction import Direction


class NeighbourGraph:
    """
    A NeighbourGraph is a class that defines the  graphical representation of the adjacency relationship between CellBoxes in the Mesh.\n


    Attributes:
        neighbour_graph (dict): a dictionary that contains cellboxes ids along with their adjacent neighbours in the following form \n

               {\n
                    <CellBox id_1>: {\n
                        "1": [id_1,...,id_n],\n
                        "2": [id_1,...,id_n],\n
                        "3": [id_1,...,id_n],\n
                        "4": [id_1,...,id_n],\n
                        "-1": [id_1,...,id_n],\n
                        "-2": [id_1,...,id_n],\n
                        "-3": [id_1,...,id_n],\n
                        "-4": [id_1,...,id_n],\n
                    },\n
                    ...,\n
                    {\n
                        <CellBox id_n>: {\n
                            ...\n
                        }\n
                    }\n
               }

    """

    def __init__(self, cellboxes=None, grid_width=0):
        # initialize graph with an empty one
        self.neighbour_graph = {}
        if cellboxes is None:
            cellboxes = []
        self.initialise_neighbour_graph(cellboxes, grid_width)
        self._is_global_mesh = False

    @classmethod
    def from_json(cls, ng_json):
        ''' 
            method that initializes a graph from a json object
            Args: 
            ng_json (json_object): json object that contains the neighbour_graph data
        '''
        neighbour_graph = {}
        for key in ng_json:
            neighbour_graph[key] = ng_json[key]
        obj = NeighbourGraph()
        obj.neighbour_graph = neighbour_graph
        return obj

    def get_graph(self):
        """
        returns the graph dict
        """
        return self.neighbour_graph

    def update_neighbour(self, index, direction, neighbours):
        """
        updates the neighbour in a certain direction
        """
        self.neighbour_graph[index][direction] = neighbours

    def add_neighbour(self, index, direction, neighbour_indx):
        """
        adds a neighbour in a certain direction

        Args:
        index (int): the index of the cellbox to be updated
        direction (int): the direction into which the neighbour will be added
        neighbour_indx (int): the index of the cellbox to be added as a neighbour
        """
        self.neighbour_graph[index][direction].append (neighbour_indx)

    def remove_node_and_update_neighbours(self, cellbox_index):
        ''' 
            method that removes a node in the neighbour_graph at a given index. remove_node_from_neighbours should be called first.
            Args: 
            cellbox_index (int): the index of the cellbox that will get removed from the neighbour_graph
        '''
        # go through all the neighbours in all the directions to remove the give cellbox_index from their neighbour_map
        direction_obj = Direction()
        for direction in direction_obj.__dict__.values():
            self.remove_node_from_neighbours(cellbox_index, direction)

        self.neighbour_graph.pop(cellbox_index)

    def get_neighbours(self, cellbox_indx, direction):
        """
        returns neighbour in a certain direction
        """
        return self.neighbour_graph[cellbox_indx][direction]

    def add_node(self, index, neighbour_map):
        '''
            method that adds a node to the neighbour_graph at a given index
            Args:
            index(int):the index at which the node will be added in the neighbour_graph
            neighbour_map (dict): a dict that contains the neighbours of the node
        '''
        self.neighbour_graph[index] = neighbour_map

    def remove_node(self, cellbox_index):
        '''
            method that removes a node to the neighbour_graph at a given index
            Args: 
            cellbox_index (int): the index of the cellbox that will get removed from the neoghbpur_graph
        '''
        self.neighbour_graph.pop(cellbox_index)

    def update_neighbours(self, cellbox_indx, new_neighbours_indx, direction, cellboxes):
        '''
            method that updates the neighbour of a certain cellbox in a specific direction. It removes cellbox_indx from the neighbour_map of its neighbours in a specific direction and add new_neighbour_indx

            Args: 
                cellbox_index (int): index of the cellbox that its neighbour will be updated
                new_neighbour_indx (int): the index of the new neighbour that will replace cellbox_index
                direction (int): an int that represents the direction of the neighbours that will get updated (e.g. north, south ,..)
                cellboxes(list<CellBox>): the list that contains all the cellboxes of the mesh

        '''
        self.remove_node_from_neighbours(cellbox_indx, direction)
        neighbour_indx_list = self.neighbour_graph[cellbox_indx][direction]
        for indx in neighbour_indx_list:
            crossing_case = self.get_neighbour_case(cellboxes[indx],
                                                    cellboxes[new_neighbours_indx[0]])

            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(
                    new_neighbours_indx[0])

            crossing_case = self.get_neighbour_case(cellboxes[indx],
                                                    cellboxes[new_neighbours_indx[1]])

            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(
                    new_neighbours_indx[1])

    def remove_node_from_neighbours(self, cellbox_indx, direction):
        '''
            method that goes through neighbours in a given direction and remove cellbox_index from their neighbour_maps
            Args:
            cellbox_indx (int): the index of the cellbox that we will go through its neighbours and remove this index from their neighbour_map
            direction (int): an int that represents the direction of the neighbours that will get updated (e.g. north, south ,..)
        '''

        # Try with int keys first as in mesh construction then try with str keys for json mesh
        try:
            neighbour_indx_list = self.neighbour_graph[cellbox_indx][direction]
        except KeyError:
            neighbour_indx_list = self.neighbour_graph[cellbox_indx][str(direction)]

        for indx in neighbour_indx_list:
            try:
                self.neighbour_graph[indx][-1*direction].remove(cellbox_indx)
            except KeyError:
                self.neighbour_graph[str(indx)][str(-1*direction)].remove(int(cellbox_indx))

    def update_corner_neighbours(self, cellbox_indx, north_west_indx, north_east_indx, south_west_indx, south_east_indx):
        '''
            method that updates the corner neighbours of cellbox_indx with the given indeces
        '''
        north_east_corner_indx = self.neighbour_graph[cellbox_indx][Direction.north_east]
        if len(north_east_corner_indx) > 0:
            self.neighbour_graph[north_east_corner_indx[0]][Direction.south_west] = [north_east_indx]

        north_west_corner_indx = self.neighbour_graph[cellbox_indx][Direction.north_west]
        if len(north_west_corner_indx) > 0:
            self.neighbour_graph[north_west_corner_indx[0]][Direction.south_east] = [north_west_indx]

        south_east_corner_indx = self.neighbour_graph[cellbox_indx][Direction.south_east]
        if len(south_east_corner_indx) > 0:
            self.neighbour_graph[south_east_corner_indx[0]][Direction.north_west] = [south_east_indx]

        south_west_corner_indx = self.neighbour_graph[cellbox_indx][Direction.south_west]
        if len(south_west_corner_indx) > 0:
            self.neighbour_graph[south_west_corner_indx[0]][Direction.north_east] = [south_west_indx]
    
    def get_neighbour_case(self, cellbox_a, cellbox_b):
        """
            Given two cellboxes (cellbox_a, cellbox_b) returns a case number
            representing where the two cellboxes are touching.

            Args:
                cellbox_a (CellBox): starting CellBox
                cellbox_b (CellBox): destination CellBox

            Returns:
                int: an int representing the direction of the adjacency between input cellbox_a and cellbox_b. The meaning of each case is as follows -

                        case 0 -> CellBoxes are not neighbours

                        case 1 -> cellbox_b is the North-East corner of cellbox_a\n
                        case 2 -> cellbox_b is East of cellbox_a\n
                        case 3 -> cellbox_b is the South-East corner of cellbox_a\n
                        case 4 -> cellbox_b is South of cellbox_a\n
                        case -1 -> cellbox_b is the South-West corner of cellbox_a\n
                        case -2 -> cellbox_b is West of cellbox_a\n
                        case -3 -> cellbox_b is the North-West corner of cellbox_a\n
                        case -4 -> cellbox_b is North of cellbox_a\n
        """
   
        long_a = cellbox_a.get_bounds().get_long_min()
        lat_a = cellbox_a.get_bounds().get_lat_min()
        long_b = cellbox_b.get_bounds().get_long_min()
        lat_b = cellbox_b.get_bounds().get_lat_min()
        def on_global_bound(cellbox_a , cellbox_b):
            """
            Given two cellboxes (cellbox_a, cellbox_b) returns a boolean
            representing whether the two cellboxes are touching on the global bound (-180,180).

            Args:
                cellbox_a (CellBox): starting CellBox
                cellbox_b (CellBox): destination CellBox

            Returns:
                bool: a boolean representing if the two cellboxes are touching on the global bound (-180,180).            
            """
            return long_a == -180  and cellbox_b.get_bounds().get_long_max() == 180 or long_b == -180 and cellbox_a.get_bounds().get_long_max() == 180 
        if self.is_global_mesh() and on_global_bound (cellbox_a , cellbox_b) :
            return self.get_global_mesh_neighbour_case(cellbox_a, cellbox_b)
        if (long_a + cellbox_a.get_bounds().get_width()) == long_b and (
                lat_a + cellbox_a.get_bounds().get_height()) == lat_b:
            return Direction.north_east
        if (long_a + cellbox_a.get_bounds().get_width() == long_b) and (
                lat_b < (lat_a + cellbox_a.get_bounds().get_height())) and (
                (lat_b + cellbox_b.get_bounds().get_height()) > lat_a):
            return Direction.east
        if (long_a + cellbox_a.get_bounds().get_width()) == long_b and (
                lat_a == lat_b + cellbox_b.get_bounds().get_height()):
            return Direction.south_east
        if ((lat_b + cellbox_b.get_bounds().get_height()) == lat_a) and (
                (long_b + cellbox_b.get_bounds().get_width()) > long_a) and (
                long_b < (long_a + cellbox_a.get_bounds().get_width())):
            return Direction.south
        if long_a == (long_b + cellbox_b.get_bounds().get_width()) and lat_a == (
                lat_b + cellbox_b.get_bounds().get_height()):
            return Direction.south_west
        if (long_b + cellbox_b.get_bounds().get_width() == long_a) and (
                lat_b < (lat_a + cellbox_a.get_bounds().get_height())) and (
                (lat_b + cellbox_b.get_bounds().get_height()) > lat_a):
            return Direction.west
        if long_a == (long_b + cellbox_b.get_bounds().get_width()) and (
                lat_a + cellbox_a.get_bounds().get_height() == lat_b):
            return Direction.north_west
        if (lat_b == (lat_a + cellbox_a.get_bounds().get_height())) and (
                (long_b + cellbox_b.get_bounds().get_width()) > long_a) and (
                long_b < (long_a + cellbox_a.get_bounds().get_width())):
            return Direction.north
        return 0  # Cells are not neighbours.
    
    def get_global_mesh_neighbour_case(self, cellbox_a, cellbox_b):
        """
            Given two cellboxes in a *global mesh* (cellbox_a, cellbox_b) returns a case number
            representing where the two cellboxes are touching.

            Args:
                cellbox_a (CellBox): starting CellBox
                cellbox_b (CellBox): destination CellBox

            Returns:
                int: an int representing the direction of the adjacency between input cellbox_a and cellbox_b. The meaning of each case is as follows -

                        case 0 -> CellBoxes are not neighbours

                        case 1 -> cellbox_b is the North-East corner of cellbox_a\n
                        case 2 -> cellbox_b is East of cellbox_a\n
                        case 3 -> cellbox_b is the South-East corner of cellbox_a\n
                        case 4 -> cellbox_b is South of cellbox_a\n
                        case -1 -> cellbox_b is the South-West corner of cellbox_a\n
                        case -2 -> cellbox_b is West of cellbox_a\n
                        case -3 -> cellbox_b is the North-West corner of cellbox_a\n
                        case -4 -> cellbox_b is North of cellbox_a\n
        """
        long_a = cellbox_a.get_bounds().get_long_min()
        lat_a = cellbox_a.get_bounds().get_lat_min()
        long_b = cellbox_b.get_bounds().get_long_min()
        lat_b = cellbox_b.get_bounds().get_lat_min()
        if (long_a + cellbox_a.get_bounds().get_width()) == abs(long_b)and (
                lat_a + cellbox_a.get_bounds().get_height()) == lat_b:
            return Direction.north_east
        if (long_a + cellbox_a.get_bounds().get_width() == abs(long_b)) and (
                lat_b < (lat_a + cellbox_a.get_bounds().get_height())) and (
                (lat_b + cellbox_b.get_bounds().get_height()) > lat_a):
            return Direction.east
        if (long_a + cellbox_a.get_bounds().get_width()) == abs(long_b) and (
                lat_a == lat_b + cellbox_b.get_bounds().get_height()):
            return Direction.south_east
        if ((lat_b + cellbox_b.get_bounds().get_height()) == lat_a) and (
                (long_b + cellbox_b.get_bounds().get_width()) > long_a) and (
                long_b < (long_a + cellbox_a.get_bounds().get_width())):
            return Direction.south
        if abs(long_a) == (long_b + cellbox_b.get_bounds().get_width()) and lat_a == (
                lat_b + cellbox_b.get_bounds().get_height()):
            return Direction.south_west
        if (long_b + cellbox_b.get_bounds().get_width() == abs(long_a)) and (
                lat_b < (lat_a + cellbox_a.get_bounds().get_height())) and (
                (lat_b + cellbox_b.get_bounds().get_height()) > lat_a):
            return Direction.west
        if abs(long_a) == (long_b + cellbox_b.get_bounds().get_width()) and (
                lat_a + cellbox_a.get_bounds().get_height() == lat_b):
            return Direction.north_west
        if (lat_b == (lat_a + cellbox_a.get_bounds().get_height())) and (
                (long_b + cellbox_b.get_bounds().get_width()) > long_a) and (
                long_b < (long_a + cellbox_a.get_bounds().get_width())):
            return Direction.north
        return 0  # Cells are not neighbours.

    def remove_neighbour(self,  index,  direction, neighbour_index):
        """ 
            remove certain neighbour in a specific direction
        """
        self.neighbour_graph[index][direction].remove(neighbour_index)

    def initialise_neighbour_graph(self, cellboxes, grid_width):
        """
         initialize the neighbour graph
        """
        for cellbox in cellboxes:
            cellbox_indx = cellboxes.index(cellbox)
            neighbour_map = self.initialise_map(
                cellbox_indx, grid_width, len(cellboxes))

            self.add_node(cellbox_indx, neighbour_map)

    def initialise_map(self, cellbox_indx, grid_width, cellboxes_length):
        """
            initialse the neighbour map of a cellbox with the given cellbox_index
        """
        neighbour_map = {Direction.north_east: [],
                         Direction.east: [],
                         Direction.south_east: [],
                         Direction.south: [],
                         Direction.south_west: [],
                         Direction.west: [],
                         Direction.north_west: [], 
                         Direction.north: []}

        # add east neighbours to neighbour graph
        if (cellbox_indx + 1) % grid_width != 0:
            neighbour_map[Direction.east].append(cellbox_indx + 1)
            # south-east neighbours
            if cellbox_indx + grid_width < cellboxes_length:
                neighbour_map[Direction.north_east].append(
                    int((cellbox_indx + grid_width) + 1))
            # north-east neighbours
            if cellbox_indx - grid_width >= 0:
                neighbour_map[Direction.south_east].append(
                    int((cellbox_indx - grid_width) + 1))

        # add west neighbours to neighbour graph
        if cellbox_indx % grid_width != 0:
            neighbour_map[Direction.west].append(cellbox_indx - 1)
            # add south-west neighbours to neighbour graph
            if cellbox_indx + grid_width < cellboxes_length:
                neighbour_map[Direction.north_west].append(
                    int((cellbox_indx + grid_width) - 1))
            # add north-west neighbours to neighbour graph
            if cellbox_indx - grid_width >= 0:
                neighbour_map[Direction.south_west].append(
                    int((cellbox_indx - grid_width) - 1))

        # add south neighbours to neighbour graph
        if cellbox_indx + grid_width < cellboxes_length:
            neighbour_map[Direction.north].append(int(cellbox_indx + grid_width))

        # add north neighbours to neighbour graph
        if cellbox_indx - grid_width >= 0:
            neighbour_map[Direction.south].append(
                int(cellbox_indx - grid_width))
        return neighbour_map
    
    def set_global_mesh (self, is_global):
        self._is_global_mesh = is_global

    def is_global_mesh (self):
        return self._is_global_mesh
    
    def get_neighbour_map (self, _id):
        return self.neighbour_graph [_id]


import logging
import math
import numpy as np
import matplotlib.pyplot as plt

import json

from meshiphi.mesh_generation.cellbox import CellBox


class Mesh:
    """
        Attributes:
            cellboxes (list<(CellBox)>): A list of CellBox objects forming the Mesh

            neighbour_graph (NeighbourGraph): A graphical representation of the adjacency
                relationship between CellBoxes in the Mesh. The neighbour_graph object conatin a dict
                of the form

               {\n
                    <CellBox id_1>: {\n
                        "1": [id_1,...,id_n],\n
                        "2": [id_1,...,id_n],\n
                        "3": [id_1,...,id_n],\n
                        "4": [id_1,...,id_n],\n
                        "-1": [id_1,...,id_n],\n
                        "-2": [id_1,...,id_n],\n
                        "-3": [id_1,...,id_n],\n
                        "-4": [id_1,...,id_n],\n
                    },\n
                    ...,\n
                    {\n
                        <CellBox id_n>: {\n
                            ...\n
                        }\n
                    }\n
               }

    """

    def __init__(self, boundary , cellboxes , neighbour_graph , max_split_depth):
        """
            Constructs a Mesh object from a given parameters.

            Args:
                boundary (Boundary): the geo-spatial/temporal boundaries (longtitude, latitude and time) of the Mesh \n
                cellboxes (list <CellBox>): a list of the cellboxes contained in the Mesh \n
                neghbour_graph (NeighbourGraph): a graphical representation of the adjacency \n
                relationship between CellBoxes in the Mesh \n      
        """
        self.boundary = boundary
        self.cellboxes = cellboxes
        self.neighbour_graph = neighbour_graph
        self.max_split_depth = max_split_depth
        self.config = {}

    # Functions for adding data to the Mesh

    def add_data_points(self, data_points):
        """
            takes a dataframe containing geospatial-temporal located values and assigns them to
            cellboxes within this Mesh.

            Args:
                data_points (DataFrame): a dataframe of datapoints to be added to the Mesh.
                    data_points is of the form \n
                    lat | long | (time)* | value_1 | ... | value_n
        """
        for cell_box in self.cellboxes:
            if isinstance(cell_box, CellBox):
                long_loc = data_points.loc[(data_points['long'] > cell_box.long) &
                                           (data_points['long'] <= (cell_box.long + cell_box.width))]
                lat_long_loc = long_loc.loc[(long_loc['lat'] > cell_box.lat) &
                                            (long_loc['lat'] <= (cell_box.lat + cell_box.height))]

                cell_box.add_data_points(lat_long_loc)

    # Functions for outputting the Mesh

    def set_config(self, config):
        self.config = config

    def get_config (self):
        return self.config

    def get_cellboxes (self):
        return self.cellboxes

    def set_cellboxes (self, cellboxes):
        self.cellboxes = cellboxes

    def get_cellboxes(self):
        """
            returns a list of dictionaries containing information about each cellbox
            in this Mesh.
            all cellboxes will include id, geometry, cx, cy, dcx, dcy

            Returns:
                cellboxes (list<dict>): a list of CellBoxes which form the Mesh.
                    CellBoxes are of the form -

                    {
                        "id": (string) ... \n
                        "geometry": (string) POLYGON(...), \n
                        "cx": (float) ..., \n
                        "cy": (float) ..., \n
                        "dcx": (float) ..., \n
                        "dcy": (float) ..., \n
                        \n
                        "value_1": (float) ..., \n
                        ..., \n
                        "value_n": (float) ... \n
                    }
        """
        return_cellboxes = []
        for cellbox in self.cellboxes:
            if isinstance(cellbox, CellBox): #checking to filter out the cellbox that was splitted and replaced

                # Get json for CellBox
                #cell = cellbox.to_json()
                # Append ID to CellBox
                #cell['id'] = str(self.cellboxes.index(cellbox))

                return_cellboxes.append(cellbox)
        return return_cellboxes

    def get_cellbox(self, long, lat):
        """
            Returns the CellBox which contains a point, given by parameters lat, long

            Args:
                long (long): longitude of a given point
                lat (float): latitude of given point

            Returns:
                cellbox (CellBox): the cellbox which contains the point given my parameters
                (long, lat)
        """
        selected_cell = []
        for cellbox in self.cellboxes:
            if isinstance(cellbox, CellBox):
                if cellbox.contains_point(lat, long):
                    selected_cell.append(cellbox)
        return selected_cell[0]

    def get_bounds(self): 
        return self.boundary
  
    def get_max_split_depth (self):
        return self.max_split_depth
    



from meshiphi.dataloaders.lut.abstract_lut import LutDataLoader
from meshiphi.mesh_generation.boundary import Boundary

import pandas as pd
from shapely.ops import unary_union
from shapely import wkt
import numpy as np

import logging

class LutCSV(LutDataLoader):    
    
    def import_data(self, bounds):
        """
        Import a list of .csv files, assign regions a value specified in
        config params, regions outside this are numpy nan values.
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            exclusion_df (pd.DataFrame): 
                Dataframe of polygons with value specified in config.
                DataFrame has columns 'geometry' and 
                data_name (read from CSV by default)
        """
        # Read in all files and create dataframe from them
        df_list = [pd.read_csv(file, index_col=False) for file in self.files]
        csv_df = pd.concat(df_list, ignore_index=True)
        
        # Make sure .csv is well formed
        assert ('geometry' in csv_df.columns), \
                "'geometry' column required"
        assert (csv_df.geometry.str.contains('POLYGON').all()), \
                "Only 'Polygon' or 'MultiPolygon' geometry allowed"
        assert (len(csv_df.columns) == 2), \
                "Dataloader only accepts .csv with 2 columns, 'geometry' and {{data_name}}"
        
        # Set data name to column in CSV
        self.data_name = list(set(csv_df.columns.values) - set(['geometry']))[0]
        # Convert strings to shapely geometries
        csv_df['geometry'] = csv_df['geometry'].apply(wkt.loads)
        # Create boundary denoting the world 
        world_polygon = Boundary([-90, 90], [-180, 180]).to_polygon()
        # Subtract out all regions with defined values
        defined_polygon = unary_union(csv_df.geometry)
        undefined_polygon = world_polygon - defined_polygon
        # Set remainder to have value np.nan
        csv_df.loc[len(csv_df.index)] = [undefined_polygon, np.nan]
        # Limit to boundary
        csv_df = self.trim_datapoints(bounds, data=csv_df)

        return csv_df


from meshiphi.dataloaders.lut.abstract_lut import LutDataLoader
from meshiphi.mesh_generation.boundary import Boundary
import logging

import pandas as pd
from shapely import wkt


# Mapping of month number to season per hemisphere
northern_seasons = {
    1: 'wi',  2: 'wi', 12: 'wi',
    3: 'sp',  4: 'sp',  5: 'sp', 
    6: 'su',  7: 'su',  8: 'su', 
    9: 'au', 10: 'au', 11: 'au',
    }
southern_seasons = {
    1: 'su',  2: 'su', 12: 'su',
    3: 'au',  4: 'au',  5: 'au', 
    6: 'wi',  7: 'wi',  8: 'wi', 
    9: 'sp', 10: 'sp', 11: 'sp',
    }

class ThicknessDataLoader(LutDataLoader):
    
    class Region:
        """
        Data storage object with region boundary as polygon, seasonal 
        sea ice densities, and a dict mapping from integer month to season
        """
        def __init__(self, name, geometry, value_dict, seasons=southern_seasons):
            self.name = name
            self.geometry = geometry
            self.value_dict = value_dict
            self.month_to_season = seasons
        
        def get_value(self, month):
            """
            Returns the sea ice density for a given month, taking into account
            seasons per hemisphere

            Args:
                month (int): Month as an integer (1 = Jan ... 12 = Dec)

            Returns:
                float: Sea Ice Density in the specified season
            """
            return self.value_dict[self.month_to_season[month]]
        
    
    def import_data(self, bounds):
        """
        Creates a simulated dataset of sea ice thickness based on 
        scientific literature.
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            thickness_df (pd.DataFrame): 
                Sea Ice thickness dataset within limits of bounds. 
                DataFrame has coordinates 'date', 'geometry', 
                and variable 'thickness'
        """
        regions = [
            self.Region('Ross E', 
                        Boundary([-90, 0], [-180, -130]).to_polygon(),
                        {'wi': 0.72, 'sp': 0.67, 'su': 1.32, 'au': 0.82, 'y': 1.07}),
            self.Region('Bellinghausen',
                        Boundary([-90, 0], [-130, -60]).to_polygon(),
                        {'wi': 0.65, 'sp': 0.79, 'su': 2.14, 'au': 0.79, 'y': 0.90}),
            self.Region('Weddell W',
                        Boundary([-90, 0], [-60, -45]).to_polygon(),
                        {'wi': 1.33, 'sp': 1.33, 'su': 1.20, 'au': 1.38, 'y': 1.33}),
            self.Region('Weddell E',
                        Boundary([-90, 0], [-45, 20]).to_polygon(),
                        {'wi': 0.54, 'sp': 0.89, 'su': 0.87, 'au': 0.44, 'y': 0.73}),
            self.Region('Indian',
                        Boundary([-90, 0], [20, 90]).to_polygon(),
                        {'wi': 0.59, 'sp': 0.78, 'su': 1.05, 'au': 0.45, 'y': 0.68}),
            self.Region('West Pacific',
                        Boundary([-90, 0], [90, 160]).to_polygon(),
                        {'wi': 0.72, 'sp': 0.68, 'su': 1.17, 'au': 0.75, 'y': 0.79}),
            self.Region('Ross W', 
                        Boundary([-90, 0], [160,   180]).to_polygon(),
                        {'wi': 0.72, 'sp': 0.67, 'su': 1.32, 'au': 0.82, 'y': 1.07}),
            self.Region('None',
                        Boundary([0, 90], [-180, 180]).to_polygon(),
                        {'wi': 0.72, 'sp': 0.67, 'su': 1.32, 'au': 0.82, 'y': 1.07},
                        seasons=northern_seasons),
        ]
        
        # Create shape to intersect regions with
        bounds_polygon = bounds.to_polygon()
        
        # For every date in range, create a new shape and value pair
        dates = pd.date_range(start=bounds.get_time_min(), 
                              end=bounds.get_time_max())
        
        thickness_df = pd.concat([
                pd.DataFrame({'time': dates,
                              'geometry': region.geometry & bounds_polygon,
                              'thickness': region.get_value(month)})
                for region in regions for month  in dates.month
            ])
        
        thickness_df = thickness_df.set_index('time').sort_index()
        
        return thickness_df


import json
import jsonschema
import logging
import datetime
import re
import pandas as pd

from meshiphi.config_validation.mesh_schema import mesh_schema
from meshiphi.config_validation.vessel_schema import vessel_schema
from meshiphi.config_validation.route_schema import route_schema
from meshiphi.config_validation.waypoints_schema import waypoints_columns

def flexi_json_input(config):
    """
    Allows flexible inputs. If a string is parsed, then assume it's a file path
    and read in as a json. If a dict is parsed, then assume it's already a 
    valid loaded json, and return it as is

    Args:
        config (str or dict): Input to translate into a dict.

    Raises:
        TypeError: If input is neither a str nor a dict, then wrong input type

    Returns:
        dict: Dictionary read from JSON
    """
    if type(config) is str:
        # If str, assume filename
        with open(config, 'r') as fp:
            config_json = json.load(fp)
    elif type(config) is dict:
        # If dict, assume it's the config
        config_json = config
    else:
        # Otherwise, can't deal with it
        raise TypeError(f"Expected 'str' or 'dict', instead got '{type(config)}'")
    
    return config_json

def validate_mesh_config(config):
    """
    Validates a mesh config

    Args:
        config (str or dict): 
            Mesh config to be validated. 
            If type 'str', tries to read in as a filename and open file as json
            If type 'dict', assumes it's already read in from a json file

    Raises:
        TypeError: Incorrect config parsed in. Must be 'str' or 'dict'
        FileNotFoundError: Could not read in file if 'str' parsed
        ValidationError: Malformed mesh config
    """
    
    def assert_valid_time(time_str):
        """
        Checks if the time strings in the config are correctly formatted

        Args:
            time_str (str): 
                String from config. Expects 'YYYY-MM-DD' or 'TODAY +- n'
        Raises:
            ValueError: String not in a valid date format
        """
        correctly_formatted = False
        # If relative time is parsed
        if re.match('TODAY[+,-]\d+', time_str.replace(' ','')):
            correctly_formatted = True
        elif time_str == 'TODAY':
            correctly_formatted = True
        # Otherwise check if date is parsed correctly
        else:
            try:
                # Checks if formatted as YYYY-MM-DD with a valid date
                datetime.date.fromisoformat(time_str)
                # If so, then it's correct
                correctly_formatted = True
            except ValueError:
                # Otherwise, keep correctly_formatted = False
                pass
        # If it failed to pass
        if not correctly_formatted:
            raise ValueError(f'{time_str} is not a valid date format!')
    
    def assert_valid_cellsize(bound_min, bound_max, cell_size):
        """
        Ensures that the initial cellbox size can evenly divide the initial
        boundary.

        Args:
            bound_min (float): Minimum value of boundary in one axis
            bound_max (float): Maximum value of boundary in the same axis
            cell_size (float): Initial cellbox size in the same axis
        """
        assert((bound_max - bound_min)%cell_size == 0), \
            f"{bound_max}-{bound_min}={bound_max-bound_min} is not evenly "+\
            f"divided by {cell_size}"
        
        
    # Deals with flexible input
    config_json = flexi_json_input(config)
    # Validate against the schema to check syntax is correct
    jsonschema.validate(instance=config_json, schema=mesh_schema)
    
    # Check that the dates in the region are valid
    assert_valid_time(config['region']['start_time'])
    assert_valid_time(config['region']['end_time'])
    
    # Check that cellbox width and height evenly divide boundary
    assert_valid_cellsize(config['region']['lat_min'],
                          config['region']['lat_max'],
                          config['region']['cell_height'])
    assert_valid_cellsize(config['region']['long_min'],
                          config['region']['long_max'],
                          config['region']['cell_width'])
    
def validate_vessel_config(config):
    """
    Validates a vessel config

    Args:
        config (str or dict):
            Vessel config to be validated.
            If type 'str', tries to read in as a filename and open file as json
            If type 'dict', assumes it's already read in from a json file

    Raises:
        TypeError: Incorrect config parsed in. Must be 'str' or 'dict'
        FileNotFoundError: Could not read in file if 'str' parsed
        ValidationError: Malformed vessel config

    """
    # Deals with flexible input
    config_json = flexi_json_input(config)
    # Validate against schema
    jsonschema.validate(instance=config_json, schema=vessel_schema)

def validate_route_config(config):
    """
    Validates a route config

    Args:
        config (str or dict):
            route config to be validated.
            If type 'str', tries to read in as a filename and open file as json
            If type 'dict', assumes it's already read in from a json file

    Raises:
        TypeError: Incorrect config parsed in. Must be 'str' or 'dict'
        FileNotFoundError: Could not read in file if 'str' parsed
        ValidationError: Malformed route config

    """
    # Deals with flexible input
    config_json = flexi_json_input(config)
    # Validate against schema
    jsonschema.validate(instance=config_json, schema=route_schema)

def validate_waypoints(waypoints):
    """_summary_

    Args:
        waypoints (str or pd.DataFrame): _description_

    Raises:
        TypeError: Incorrect config parsed in. Must be 'str' or 'pd.DataFrame' 
        FileNotFoundError: Could not read in file if 'str' parsed
        AssertionError: Malformed waypoints config
    """

    if type(waypoints) is str:
        # If str, assume filename
        waypoints_df = pd.read_csv(waypoints).reset_index()
    elif type(waypoints) is pd.core.frame.DataFrame:
        # If dataframe, assume it's the loaded waypoints
        waypoints_df = waypoints.reset_index()
    else:
        # Otherwise, can't deal with it
        raise TypeError(
            f"Expected 'str' or 'dict', instead got '{type(waypoints)}'"
            )
    # Assert that all the required columns exist
    assert(all(col in waypoints_df.columns for col in waypoints_columns)), \
        f'Expected the following columns to exist: {waypoints_columns}'
    
    # Assert that at least one source and destination waypoint selected
    assert('X' in set(waypoints_df['Source'])), \
        'No source waypoint defined!'
    assert('X' in set(waypoints_df['Destination'])), \
        'No destination waypoint defined!'
        
    # Assert only numeric values given as coordinates
    assert(pd.to_numeric(waypoints_df['Lat'], errors='coerce').notnull().all()), \
        'Non-numeric value in "Lat" column of waypoints'
    assert(pd.to_numeric(waypoints_df['Long'], errors='coerce').notnull().all()), \
        'Non-numeric value in "Long" column of waypoints'




from meshiphi.dataloaders.vector.abstract_vector import VectorDataLoader
from meshiphi.utils import gaussian_random_field

import logging
import pandas as pd
import numpy as np

class VectorGRFDataLoader(VectorDataLoader):

    def add_default_params(self, params):
        '''
        Set default values for abstract GRF dataloaders, starting by
        including defaults for scalar dataloaders.
        
        Args:
            params (dict): 
                Dictionary containing attributes that are required for the
                shape being loaded. Must include 'shape'.
            
        Returns:
            (dict): 
                Dictionary of attributes the dataloader will require, 
                completed with default values if not provided in config.
        '''
        # Set default vector dataloader params
        params = super().add_default_params(params)
        
        # Params that all GRF dataloaders need
        if 'seed' not in params:
            params['seed'] = None
        if 'size' not in params:
            params['size'] = 512
        if 'alpha' not in params:
            params['alpha'] = 3
        # Column/variable names
        if params['data_name'] is None:
            params['data_name'] = 'uC,vC'
        if 'vec_x' not in params:
            params['vec_x'] = params['data_name'].split(',')[0]
        if 'vec_y' not in params:
            params['vec_y'] = params['data_name'].split(',')[1]
        # Min/Max magnitude
        if 'min' not in params:
            params['min'] = 0
        if 'max' not in params:
            params['max'] = 10

        return params
        
    def import_data(self, bounds):
        '''
        Creates data in the form of a Gaussian Random Field
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            pd.DataFrame: 
                Scalar dataset within limits of bounds. Dataset has coordinates
                'lat', 'long', and variable name 'data'. \n
                NOTE - In __init__, variable name will be set to data_name if 
                defined in config
                
        '''

        def grf_to_vector(magnitudes, directions, min_val, max_val):
            # Scale to max/min
            magnitudes = magnitudes * (max_val - min_val) + min_val
            
            dy = np.cos(directions) * magnitudes
            dx = np.sin(directions) * magnitudes
            
            return dx, dy
        
        # Set seed for generation. If not specified, will be 'random'
        np.random.seed(self.seed)
        
        # Create a GRF of magnitudes and angles
        magnitudes = gaussian_random_field(self.size, self.alpha)
        directions = gaussian_random_field(self.size, self.alpha)
        directions = np.radians(360*directions)
        
        vec_x, vec_y = grf_to_vector(magnitudes, directions, self.min, self.max)
        
    
        # Set up domain of field
        lat_array = np.linspace(bounds.get_lat_min(), 
                                bounds.get_lat_max(), 
                                self.size)
        long_array = np.linspace(bounds.get_long_min(), 
                                 bounds.get_long_max(), 
                                 self.size)
        latv, longv = np.meshgrid(lat_array, long_array, indexing='ij')

        # Create an entry for each row in final dataframe
        rows = [
            {'lat': latv[i,j], 'long': longv[i,j], 
             self.vec_x: vec_x[i,j], self.vec_y: vec_y[i,j]}
            for i in range(self.size) 
            for j in range(self.size)
            ]
        # Cast to dataframe
        data = pd.DataFrame(rows).set_index(['lat','long'])
        # Set to xarray dataset
        data = data.to_xarray()
        
        return data



from meshiphi.mesh_generation.boundary import Boundary
from meshiphi.mesh_generation.jgrid_aggregated_cellbox import JGridAggregatedCellBox
from meshiphi.mesh_generation.cellbox import CellBox
import numpy as np


class JGridCellBox (CellBox):
    """
    A JGridCellBox represnts a subclass of CellBox tailored to guarantee compatability with the earlier Java implemnation (which is based on coordinates and focus).


    Attributes:
        Bounds (Boundary): object that contains the latitude and logtitute range and the time range

    Note:
        All geospatial boundaries of a CellBox are given in a 'EPSG:4326' projection
    """

    def __init__(self, bounds, id):
        """

            Args:
                bounds(Boundary): encapsulates latitude, longtitude and time range of the CellBox
        """
        # Box information relative to bottom left
        self.bounds = bounds
        self.parent = None
        self.minimum_datapoints = 3000
        self.split_depth = 0
        self.data_source = None
        self.id = id
        self.x_coord = 0
        self.y_coord = 0
        self.focus = []
        self.land_locked = False
        self.initial_bounds = None

    @classmethod
    def init_from_cellbox(cls , cellbox):
        obj = JGridCellBox (cellbox.bounds , cellbox.id)
        obj.parent = cellbox.parent
        obj.split_depth = cellbox.split_depth
        obj.data_source = cellbox.data_source
        obj.id = cellbox.id
        return obj

    def split(self, start_id):
        """
            splits the current cellbox into 4 corners, returns as a list of cellbox objects.
            args
            start_id : represenst the start of the splitted cellboxes ids, usuallly it is the number of the existing cellboxes
            Returns:
                split_boxes (list<CellBox>): The 4 corner cellboxes generates by splitting
                    this current cellboxes and dividing the data_points contained between.
        """
        # split using the CellBox method then perform the extra JGridCellBox logic
        split_boxes = CellBox.split(self, start_id)
        jgrid_split_boxes = []
        # set CellBox split_depth, data_source and parent
        for split_box in split_boxes:
            index = split_boxes.index(split_box)
            jgrid_split_box = JGridCellBox.init_from_cellbox(split_box)
            # set gridCoord of split boxes equal to parent.
            jgrid_split_box.set_grid_coord(self.x_coord, self.y_coord)
            # create focus for split boxes.
            jgrid_split_box.set_focus(self.get_focus().copy())
            jgrid_split_box.add_to_focus(index)
            jgrid_split_box.set_initial_bounds (self.initial_bounds)
            jgrid_split_boxes.append(jgrid_split_box)
        return jgrid_split_boxes

    def aggregate(self):
        '''
            aggregates JGridCellBox data using the associated data_source's aggregate type and returns AggregatedJGridCellBox object

        '''
        agg_dict = {}
        for source in self.get_data_source():
            loader = source.get_data_loader()

            data_name = loader.data_name
            parent = self.get_parent()
            # check if the data name has many entries (ex. uC,vC)
            if ',' in data_name:
               
                agg_value = loader.get_value(self.initial_bounds)
            else:
                # get the aggregated value from the associated DataLoader
                agg_value = loader.get_value(self.bounds)
                if np.isnan(agg_value[data_name]):
                     if source.get_value_fill_type() == 'parent':
                  		# if the agg_value empty and get_value_fill_type is parent, then use the parent bounds
                        while parent is not None and np.isnan(agg_value[data_name]):
                                     agg_value = loader.get_value(
                                         parent.bounds)
                                     parent = parent.get_parent()
                     else:# not parent, so either float or Nan so set the agg_Data to value_fill_type
                         agg_value[data_name] = source.get_value_fill_type()
            if data_name == "SIC":
                number_of_points = loader.get_value (self.bounds , "COUNT")['SIC']
                agg_dict.update ({"SIC_COUNT": number_of_points})
            agg_dict.update (agg_value) # combine the aggregated values in one dict 

        agg_cellbox = JGridAggregatedCellBox (self.bounds , agg_dict , self.get_id())
   
        agg_cellbox.set_node_string(self.get_node_string())
        # print (self.get_node_string())

        return agg_cellbox 
 

    def set_grid_coord(self, xpos, ypos):
        """
            sets up initial grid-coordinate when creating a j_grid

            for use in j_grid impl
        """
        self.x_coord = xpos
        self.y_coord = ypos

    def get_focus(self):
        """
            returns the focus of this cellbox

            for use in j_grid impl
        """
        return self.focus

    def add_to_focus(self, focus):
        """
            append additional information to the focus of this cellbox
            to be used when splitting.

            for use in j_grid impl
        """
        self.focus.append(focus)

    def grid_coord(self):
        """
            returns a string representation of the grid_coord of this cellbox

            for use in j_grid regression testing
        """
        return "(" + str(int(self.x_coord)) + "," + str(int(self.y_coord)) + ")"

    def get_node_string(self):
        """
            returns a string representing the node of this cellbox


            for use in j_grid regression testing
        """
        node_string = self.grid_coord() + " F:" + str(len(self.get_focus()))

        focus_string = "["
        for focus in self.get_focus():
            focus_string += str(focus) + " "
        focus_string += "]"
        return node_string + " " + focus_string

    def set_initial_bounds(self, bounds):
        """
                returns a string representation of the grid_coord of this cellbox

                for use in j_grid regression testing
        """
        self.initial_bounds= bounds

    def is_land(self):
        """
        checks if the current cellbox is land using the loader
        """
        is_land = False
        for source in self.data_source:
            loader = source.get_data_loader()
            data_name = 'is_land'
            if loader.data_name == data_name:
                is_land = loader.get_value (self.bounnds)[data_name]  
        return is_land
    
    def set_focus(self, focus):
        """
            initialize the focus of this cellbox
            for use in j_grid regression testing
        """
        self.focus = focus

        
    def contains_point(self, lat, long):
        """
            Returns true if a given lat/long coordinate is contained within this cellbox.

            Args:
                lat (float): latitude of a given point
                long (float): longitude of a given point

            Returns:
                contains_points (bool): True if this CellBox contains a point given by
                    parameters (lat, long)
        """
        if (lat >= self.bounds.get_lat_min()) & (lat <= self.bounds.get_lat_max()):
            if (long >= self.bounds.get_long_min()) & (long <= self.bounds.get_long_max()):
                return True
        return False



from meshiphi.dataloaders.lut.abstract_lut import LutDataLoader
from meshiphi.mesh_generation.boundary import Boundary
import logging

import geopandas as gpd
import pandas as pd
from shapely.ops import unary_union

class ScotlandNCMPA(LutDataLoader):
        
    
    def import_data(self, bounds):
        """
        Creates a simulated dataset of sea ice thickness based on 
        scientific literature.
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            exclusion_df (pd.DataFrame): 
                Dataframe of exclusion zones around scotland. 
                DataFrame has coordinates 'date', 'shape', 
                and variable 'thickness'
        """
        # Read in all files specified and extract geometry of exclusion zones
        gdf_list = [gpd.read_file(file) for file in self.files]
        exclusion_df = gpd.GeoDataFrame(pd.concat(gdf_list, ignore_index=True))
        # Denote all geometries as exclusion zones
        exclusion_df['exclusion_zone'] = True
        exclusion_df = pd.DataFrame(exclusion_df[['geometry','exclusion_zone']])
        
        # Create boundary denoting the world 
        world_polygon = Boundary([-90, 90], [-180, 180]).to_polygon()
        # Subtract out all exclusion zones
        exclusion_polygon = unary_union(exclusion_df.geometry)
        inclusion_polygon = world_polygon - exclusion_polygon
        # Add line
        exclusion_df.loc[len(exclusion_df.index)] = [inclusion_polygon, False]
        # Limit to boundary
        exclusion_df = self.trim_datapoints(bounds, data=exclusion_df)

        return exclusion_df


from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader

import logging

import xarray as xr

class BSOSESeaIceDataLoader(ScalarDataLoader):
    def import_data(self, bounds):
        """
        Reads in data from a BSOSE Sea Ice NetCDF file.
        Renames coordinates to 'lat' and 'long', and renames variable to
        'SIC'

        Args:
            bounds (Boundary): Initial boundary to limit the dataset to

        Returns:
            xr.Dataset:
                BSOSE Sea Ice dataset within limits of bounds.
                Dataset has coordinates 'lat', 'long', and variable 'SIC'

        Raises:
            ValueError:
                If units specified in config,
                and value not 'fraction' or 'percentage'
        """
        # Open Dataset
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files)
        # Change column names
        data = data.rename({'SIarea': 'SIC',
                            'YC': 'lat',
                            'XC': 'long'})
        
        # Change domain of dataset from [0:360) to [-180:180)
        data = data.assign_coords(long=((data.long + 180) % 360) - 180)
        # Sort the 'long' axis so that sel() will work
        data = data.sortby('long')
        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)

        if hasattr(self, 'units'):
            logging.info(f'- Changing units of data to {self.units}')
            # Convert to percentage form if requested in params
            if self.units == 'percentage':
                data = data.assign(SIC= data['SIC'] * 100)
            elif self.units == 'fraction':
                pass # BSOSE data already in fraction form
            else:
                raise ValueError("Parameter 'units' not understood."
                                 "Expected 'percentage' or 'fraction',"
                                f"but received {self.units}")
        else:
            # Convert to percentage form by default (as expected by the vessel performance models)
            data = data.assign(SIC=data['SIC'] * 100)

        return data



from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader

import logging
import pandas as pd
import numpy as np

class ShapeDataLoader(ScalarDataLoader):
    
    def add_default_params(self, params):
        '''
        Set default values for abstract shape dataloaders, starting by
        including defaults for scalar dataloaders.
        
        Args:
            params (dict): 
                Dictionary containing attributes that are required for the
                shape being loaded. Must include 'shape'.
            
        Returns:
            (dict): 
                Dictionary of attributes the dataloader will require, 
                completed with default values if not provided in config.
        '''
        # Add default scalar params
        params = super().add_default_params(params)
        
        # Number of datapoints to populate per axis
        if 'nx' not in params:
            params['nx'] = 101
        if 'ny' not in params:
            params['ny'] = 101

        # Define default circle parameters
        if params['dataloader_name'] == 'circle':
            if 'radius' not in params:
                params['radius'] = 1
            if 'centre' not in params:
                params['centre'] = (None, None)
        # Define default square parameters
        elif params['dataloader_name'] == 'square':
            if 'side_length' not in params:
                params['side_length'] = 1
            if 'centre' not in params:
                params['centre'] = (None, None)
        # Define default gradient params
        elif params['dataloader_name'] == 'gradient':
            if 'vertical' not in params:
                params['vertical'] = True
        # Define default checkerboard params
        elif params['dataloader_name'] == 'checkerboard':
            if 'gridsize' not in params:
                params['gridsize'] = (1,1)
            
        return params
        
    
    def import_data(self, bounds):
        '''
        Generates data in the form of an abstract shape, such as circle,
        checkerboard, or gradient. This method acts like a factory in that
        it simply selects the correct shape method to enact
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            data_xr (xarray): 
                xarray with coordinates within bounds, and values between
                [0:1]. xarray has dimensions 'lat', 'long', 'time', and
                'dummy_data' (by default)
        '''
        # TODO Move self.lat/long = np.linspace here after reg tests pass

        # Generate abstract data set
        if self.dataloader_name == 'circle':
            data = self.gen_circle(bounds)
        elif self.dataloader_name == 'checkerboard':
            data = self.gen_checkerboard(bounds)
        elif self.dataloader_name == 'gradient':
            data = self.gen_gradient(bounds)
        else:
            raise ValueError(
                f'Unknown abstract shape type: {self.dataloader_name}'
                )

        data_xr = data.set_index(['lat', 'long']).to_xarray()
        # No need to trim data, as was defined by bounds

        return data_xr
    
    def gen_circle(self, bounds):
        """
            Generates a circle within bounds of lat/long min/max.
            Circle centre and radius can be defined in the config, as well as
            resolution of simulated datapoints
            Args:
                bounds (Boundary): Limits of lat/long to generate within
        """
        logging.info("\tSetting up boundary of dataset")
        # Generate rows
        self.lat  = np.linspace(bounds.get_lat_min(), bounds.get_lat_max(), self.ny)    
        # Generate cols
        self.long = np.linspace(bounds.get_long_min(),bounds.get_long_max(),self.nx)        

        # Set centre as centre of data_grid if none specified
        c_y = self.lat[int(self.ny/2)]  if not self.centre[0] else self.centre[0]
        c_x = self.long[int(self.nx/2)] if not self.centre[1] else self.centre[1]
        
        # Create vectors for row and col idx's
        y = np.vstack(np.linspace(bounds.get_lat_min(), 
                                  bounds.get_lat_max(), 
                                  self.ny))
        x = np.linspace(bounds.get_long_min(), bounds.get_long_max(), self.nx)

        logging.info("\tCreating mask of circle")
        # Create a 2D-array with distance from defined centre
        dist_from_centre = np.sqrt((x-c_x)**2 + (y-c_y)**2)
        # Turn this into a mask of values within radius
        mask = dist_from_centre <= self.radius
        # Set up empty dataframe to populate with dummy data
        dummy_df = pd.DataFrame(columns = ['lat', 'long', 'dummy_data'])
        logging.info("\tGenerating dataset")
        # For each combination of lat/long
        for i in range(self.ny):
            for j in range(self.nx):
                # Create a new row, adding mask value
                row = pd.DataFrame(data={'lat':self.lat[i], 
                                         'long':self.long[j], 
                                         'dummy_data':mask[i][j]}, index=[0])
                dummy_df = pd.concat([dummy_df, row], ignore_index=True)
                
        # Change boolean values to int
        dummy_df = dummy_df.replace(False, 0)
        dummy_df = dummy_df.replace(True, 1)

        return dummy_df

    def gen_gradient(self, bounds):
        """
            Generates a gradient within bounds of lat/long min/max.
            Gradient direction can be defined in the config, as well as
            resolution of simulated datapoints
            Args:
                bounds (Boundary): Limits of lat/long to generate within
        """
        logging.info("\tSetting up boundary of dataset")
        # Generate rows
        self.lat  = np.linspace(bounds.get_lat_min(), 
                                bounds.get_lat_max(), 
                                self.ny)    
        # Generate cols
        self.long = np.linspace(bounds.get_long_min(),
                                bounds.get_long_max(),
                                self.nx)
        
        logging.info("\tCreating gradient of values")
        #Create 1D gradient
        if self.vertical:   gradient = np.linspace(0,1,self.ny)
        else:               gradient = np.linspace(0,1,self.nx)
            
        dummy_df = pd.DataFrame(columns = ['lat', 'long', 'dummy_data'])
        logging.info("- Generating dataset")
        # For each combination of lat/long
        for i in range(self.ny):
            for j in range(self.nx):
                # Change dummy data depending on which axis to gradient
                datum = gradient[i] if self.vertical else gradient[j]
                # Create a new row, adding datum value
                row = pd.DataFrame(data={'lat':self.lat[i], 
                                         'long':self.long[j], 
                                         'dummy_data':datum}, index=[0])
                dummy_df = pd.concat([dummy_df, row], ignore_index=True)  
        
        return dummy_df

    def gen_checkerboard(self, bounds):
        """
            Generates a checkerboard pattern within bounds of lat/long min/max
            Square size can be defined in the config, as well as resolution of 
            simulated datapoints
            Args:
                bounds (Boundary): Limits of lat/long to generate within
        """
        logging.info("\tSetting up boundary of dataset")
        # Generate rows
        self.lat  = np.linspace(bounds.get_lat_min(), 
                                bounds.get_lat_max(), 
                                self.ny, endpoint=False)    
        # Generate cols
        self.long = np.linspace(bounds.get_long_min(),
                                bounds.get_long_max(),
                                self.nx, endpoint=False)

        logging.info("- Creating series of 0's and 1's for lat/long")
        # Create checkerboard pattern
        # Create horizontal stripes of 0's and 1's, stripe size defined by gridsize
        horizontal = np.floor((self.lat - bounds.get_lat_min()) \
                              / self.gridsize[1]) % 2
        # Create vertical stripes of 0's and 1's, stripe size defined by gridsize
        vertical   = np.floor((self.long - bounds.get_long_min())\
                              / self.gridsize[0]) % 2   
        dummy_df = pd.DataFrame(columns = ['lat', 'long', 'dummy_data'])
        logging.info("- Generating dataset")
        # For each combination of lat/long
        for i in range(self.ny):
            for j in range(self.nx):
                # Horizontal XOR Vertical should create boxes
                datum = (horizontal[i] + vertical[j]) % 2
                # Create a new row, adding datum value
                row = pd.DataFrame(data={'lat':self.lat[i], 
                                         'long':self.long[j], 
                                         'dummy_data':datum}, index=[0])
                dummy_df = pd.concat([dummy_df, row], ignore_index=True)
        
        return dummy_df    

    # TODO Add square



__version__ = "1.1.4"

__description__ = "CartograPhi: Earth's digital twin mapped on a non-uniform mesh"
__license__ = "MIT"
__author__ = "Samuel Hall, George Coombs, Harrison Abbot, Ayat Fekry, Jonathan Smith, Maria Fox, James Byrne, Michael Thorne"
__email__ = "polarroute@bas.ac.uk"
__copyright__ = "2022-2023, BAS AI Lab"

# Wrapped in try-except so that setup.py can import cartographi without crashing due to dependency errors
try:
    from cartographi.mesh_generation.mesh_builder import MeshBuilder as MeshBuilder
    from cartographi.dataloaders.factory import DataLoaderFactory as DataLoaderFactory
    from cartographi.mesh_generation.boundary import Boundary as Boundary

except ModuleNotFoundError as err:
    print(f'{err}\n Is CartograPhi installed correctly?')



from meshiphi.dataloaders.dataloader_interface import DataLoaderInterface
from abc import abstractmethod

from pyproj import Transformer, CRS

import logging
import numpy as np
import xarray as xr
import pandas as pd
from rasterio.enums import Resampling

from meshiphi.utils import round_to_sigfig



class ScalarDataLoader(DataLoaderInterface):
    '''
    Abstract class for all scalar Datasets.
    '''
    def __init__(self, bounds, params):
        '''
        This is where large-scale operations are performed, 
        such as importing data, downsampling, reprojecting, and renaming 
        variables
        
        Args:
            bounds (Boundary): 
                Initial mesh boundary to limit scope of data ingest
            params (dict): 
                Values needed by dataloader to initialise. Unique to each
                dataloader

        Attributes:
            self.data (pd.DataFrame or xr.Dataset): 
                Data stored by dataloader to use when called upon by the mesh.
                Must be saved in mercator projection (EPSG:4326), with 
                coordinates names 'lat', 'long', and 'time' (if applicable).
            self.data_name (str): 
                Name of scalar variable. Must be the column name if self.data
                is pd.DataFrame. Must be variable if self.data is xr.Dataset
                
        Raises:
            ValueError: If no data lies within the parsed boundary
        '''
        # Translates parameters from config input to desired inputs
        params = self.add_default_params(params)
        logging.info(f"Initialising {params['dataloader_name']} dataloader")
        # Creates a class attribute for all keys in params
        for key, val in params.items():
            setattr(self, key, val)
            
        # Read in and manipulate data to standard form
        self.data = self.import_data(bounds)
        if 'files' in params:
            logging.info('\tFiles read:')
            for file in self.files:
                logging.info(f'\t\t{file}')
        # If need to downsample data
        self.data = self.downsample()
        # If need to reproject data
        if self.in_proj != self.out_proj:
            self.data = self.reproject(
                                in_proj  = self.in_proj,
                                out_proj = self.out_proj,
                                x_col    = self.x_col,
                                y_col    = self.y_col
                                )
        # Get data name from column name if not set in params
        if self.data_name is None:
            logging.debug('\tSetting self.data_name from column name')
            self.data_name = self.get_data_col_name()
        # or if set in params, set col name to data name
        else:
            logging.debug(f'\tSetting data column name to {self.data_name}')
            self.data = self.set_data_col_name(self.data_name)

        # Calculate fraction of boundary that data covers
        data_coverage = self.calculate_coverage(bounds)
        logging.info("\tMercator data range (roughly) covers "+\
                    f"{np.round(data_coverage*100,0).astype(int)}% "+\
                     "of initial boundary")
        # If there's 0 datapoints in the initial boundary, raise ValueError
        if data_coverage == 0:
            logging.error('\tDataloader has no data in initial region!')
            raise ValueError(f"Dataloader {params['dataloader_name']}"+\
                              " contains no data within initial region!")
        else:
            # Cut dataset down to initial boundary
            logging.info(
                "\tTrimming data to initial boundary: {min} to {max}".format(
                    min=(bounds.get_lat_min(), bounds.get_long_min()),
                    max=(bounds.get_lat_max(), bounds.get_long_max())
                ))
            
            self.data = self.trim_datapoints(bounds)

    @abstractmethod
    def import_data(self, bounds):
        '''
        User defined method for importing data from files, or even generating 
        data from scratch
                
        Returns:
            xr.Dataset or pd.DataFrame:
                Coordinates and data being imported from file \n
                if xr.Dataset, 
                    - Must have coordinates 'lat' and 'long'
                    - Must have single data variable
                    
                if pd.DataFrame, 
                    - Must have columns 'lat' and 'long'
                    - Must have single data column
                    
                Downsampling and reprojecting happen in __init__() method
        '''
        pass
        
    def add_default_params(self, params):
        '''
        Set default values for all scalar dataloaders. This function should be
        overloaded to include any extra params for a specific dataloader
        
        Args:
            params (dict): 
                Dictionary containing attributes that are required for each 
                dataloader.
            
        Returns:
            (dict): 
                Dictionary of attributes the dataloader will require, 
                completed with default values if not provided in config.
        '''
        if 'dataloader_name' not in params:
            params['dataloader_name'] = self.__class__.__name__
        
        if 'data_name' not in params:
            params['data_name'] = None

        if 'downsample_factors' not in params:
            params['downsample_factors'] = [1,1]

        if 'aggregate_type' not in params: 
            params['aggregate_type']  = 'MEAN'
            
        if 'min_dp' not in params:
            params['min_dp'] = 5
            
        if 'in_proj' not in params:
            params['in_proj'] = 'EPSG:4326'
            
        if 'out_proj' not in params:
            params['out_proj'] = 'EPSG:4326'
            
        if 'x_col' not in params:
            params['x_col'] = 'lat'

        if 'y_col' not in params:
            params['y_col'] = 'long'
            
        if 'fast_reprojection' not in params:
            params['fast_reprojection'] = False
            
        return params

    def calculate_coverage(self, bounds, data=None):
        """
        Calculates percentage of boundary covered by dataset

        Args:
            bounds (Boundary): 
                Boundary being compared against
            data (pd.DataFrame or xr.Dataset): 
                Dataset with 'lat' and 'long' coordinates. 
                Extent calculated from min/max of these coordinates. 
                Defaults to objects internal dataset.
        
        Returns:
            float:
                Decimal fraction of boundary covered by the dataset
        """
        def calculate_coverage_from_df(bounds, data):
            data = data.dropna().reset_index()
            # If empty dataframe, 0% coverage
            if data.empty:
                return 0
            # If no valid coordinates within data range, 0% coverage
            elif data.lat.size == 0 or data.long.size == 0:
                return 0
            # Otherwise, calculate coverage, assuming rectangular region 
            # in mercator projection
            else:
                # Get range of latitude values
                data_lat_range = data.lat.max() - data.lat.min()
                bounds_lat_range = bounds.get_lat_max() - bounds.get_lat_min()
                # Get range of longitude values
                data_long_range = data.long.max() - data.long.min()
                bounds_long_range = bounds.get_long_max() - bounds.get_long_min()
                # Calcualte area of each region
                data_area = data_lat_range * data_long_range
                bounds_area = bounds_lat_range * bounds_long_range
                # If data area completely covers bounds, 100% coverage
                if data_area >= bounds_area:
                    return 1
                # Otherwise return decimal fraction
                else:
                    return data_area / bounds_area
                
                
        def calculate_coverage_from_xr(bounds, data):
            # Remove all NaN columns/rows
            data = data.dropna(dim="lat", how="all")
            data = data.dropna(dim="long", how="all")
            # If no valid coordinates within data range, 0% coverage
            if data.lat.size == 0 or data.long.size == 0:
                return 0
            # Otherwise, calculate coverage, assuming rectangular region 
            # in mercator projection
            else:
                # Get range of latitude values
                data_lat_range = data.lat.max().item() - data.lat.min().item()
                bounds_lat_range = bounds.get_lat_max() - bounds.get_lat_min()
                # Get range of longitude values
                data_long_range = data.long.max().item() - data.long.min().item()
                bounds_long_range = bounds.get_long_max() - bounds.get_long_min()
                # Calcualte area of each region
                data_area = data_lat_range * data_long_range
                bounds_area = bounds_lat_range * bounds_long_range
                # If data area completely covers bounds, 100% coverage
                if data_area >= bounds_area:
                    return 1
                # Otherwise return decimal fraction
                else:
                    return data_area / bounds_area
        # Use self.data if not no explicit dataset specified
        if data is None:
            data = self.data
        # Calculate data coverage fraction
        if type(self.data) == pd.core.frame.DataFrame:
            return calculate_coverage_from_df(bounds, data)
        elif type(self.data) == xr.core.dataset.Dataset:
            return calculate_coverage_from_xr(bounds, data)

    def trim_datapoints(self, bounds, data=None):
        '''
        Trims datapoints from self.data within boundary defined by 'bounds'.
        self.data can be pd.DataFrame or xr.Dataset
        
        Args:
            bounds (Boundary): Limits of lat/long/time to select data from

        Returns:
            pd.DataFrame or xr.Dataset: 
                Trimmed dataset in same format as self.data
        '''
        def trim_datapoints_from_df(data, bounds):
            '''
            Extracts datapoints from a pd.DataFrame
            
            Args:
                data (pd.DataFrame): Datapoints in dataframe
                bounds (Boundary): 
                    Maximum extent of datapoints to include. 
                    Will be cut to extent of spatial boundary 
                    (and temporal if data includes time)
                    
            Returns:
                pd.DataFrame:
                    Trimmed dataset inclusive of spatial upper bound, and 
                    exclusive of spatial lower bound. Inclusive of both
                    upper and lower time bounds
            '''
            # Mask off any positions not within spatial bounds
            mask = (data['lat']  > bounds.get_lat_min())  & \
                   (data['lat']  <= bounds.get_lat_max())  & \
                   (data['long'] > bounds.get_long_min()) & \
                   (data['long'] <= bounds.get_long_max())
            # Mask with time if time column exists
            if 'time' in data.columns:
                mask &= (data['time'] >= bounds.get_time_min()) & \
                        (data['time'] <= bounds.get_time_max())
                        
            # Return column of data from within bounds
            return data.loc[mask]

        def trim_datapoints_from_xr(data, bounds):
            '''
            Extracts datapoints from a xr.Dataset
            
            Args:
                data (xr.Dataset): Datapoints in dataframe
                bounds (Boundary): 
                    Maximum extent of datapoints to include. 
                    Will be cut to extent of spatial boundary 
                    (and temporal if data includes time)
                    
            Returns:
                xr.Dataset:
                    Trimmed dataset inclusive of spatial upper bound, and 
                    exclusive of spatial lower bound. Inclusive of both
                    upper and lower time bounds
            '''
            # Select data region within spatial bounds
            # NOTE slice in xarray is inclusive of bounds
            data = data.sel(lat=slice(bounds.get_lat_min(),  bounds.get_lat_max() ))
            data = data.sel(long=slice(bounds.get_long_min(), bounds.get_long_max()))
            # Select data region within temporal bounds if time exists as a coordinate
            if 'time' in data.coords.keys():
                data = data.sel(time=slice(bounds.get_time_min(),  bounds.get_time_max()))

            # Trim off any data on the min boundary to be consistent with df
            if bounds.get_lat_min() in data.lat:
                data = data.where(data.lat  != bounds.get_lat_min(), drop=True)
            if bounds.get_long_min() in data.long:
                data = data.where(data.long != bounds.get_long_min(), drop=True)
            
            # Return column of data from within bounds
            return data
        
        # If no specific data passed in, default to entire dataset
        if data is None:
            data = self.data
        
        # Skip trimming if data already completely within bounds
        if data.lat.min() >  bounds.get_lat_min() and \
           data.lat.max() <= bounds.get_lat_max() and \
           data.long.min() >  bounds.get_long_min() and \
           data.long.max() <= bounds.get_long_max():
            logging.debug('\tData is already trimmed to bounds!')
            return data
        
        if type(data) == pd.core.frame.DataFrame:
            return trim_datapoints_from_df(data, bounds)
        elif type(data) == xr.core.dataset.Dataset:
            return trim_datapoints_from_xr(data, bounds)

    def get_dp_from_coord(self, long=None, lat=None, return_coords=False):
        '''
        Extracts datapoint from self.data with lat and long specified in kwargs.
        self.data can be pd.DataFrame or xr.Dataset. Will return multiple values
        if one set of coordinates have multiple entries (e.g. time series data)
        
        Args:
            long (float): Longitude coordinate to search for
            lat (float) : Latitude coordinate to search for
            
        Returns:
            pd.Series:  Column of data values with chosen lat/long. Could be many 
                datapoints because either bad data or multiple time steps 
        '''
        def get_dp_from_coord_df(data, name, long, lat, return_coords):
            '''
            Extracts a singular datapoint from dataframe
            
            Args:
                data (pd.DataFrame): 
                    Dataset to search within
                name (str): 
                    Variable name to extract
                long (float): 
                    Longitude of datapoint requested
                lat (float): 
                    Latitude of datapoint requested
                return_coords (bool): 
                    Flag to choose if coords 
                    are returned with the datapoint or not
                    
            Returns:
                pd.DataFrame:  
                    Dataframe containing one row with the datapoint
                    (and coords if requested). If no datapoint found,
                    returns empty DataFrame.
            '''
            # Mask off any positions not within spatial bounds
            mask = (data['lat']  == lat)  & \
                   (data['long'] == long) 

            # Include lat/long/time if requested
            if return_coords: columns = list(data.columns)
            else:             columns = [name]
            # Return column of data from within bounds
            return data.loc[mask][columns]
        
        def get_dp_from_coord_xr(data, name, long, lat, return_coords):
            '''
            Extracts a singular datapoint from xr.Dataset
            
            Args:
                data (xr.Dataset): 
                    Dataset to search within
                name (str): 
                    Variable name to extract
                long (float): 
                    Longitude of datapoint requested
                lat (float): 
                    Latitude of datapoint requested
                return_coords (bool): 
                    Flag to choose if coords 
                    are returned with the datapoint or not
                    
            Returns:
                pd.DataFrame:  
                    Dataframe containing one row with the datapoint
                    (and coords if requested). If no datapoint found,
                    returns empty DataFrame.
            '''
            # Select data region within spatial bounds
            data = data.sel(lat=lat, long=long)
            # Cast as a pd.DataFrame
            data = data.to_dataframe().reset_index()
            # Include lat/long/time if requested
            if return_coords: columns = list(data.columns)
            else:             columns = [name]
            # Return column of data from within bounds
            return data[columns]
        
        # Ensure that lat and long provided
        assert (lat is not None) and (long) is not None, \
            'Must provide lat and long to this method!'
            
        # Choose which method to retrieve data based on input type
        if hasattr(self, 'data_name'): data_name = self.data_name
        else:                          data_name = self.get_data_col_name()
        
        if type(self.data) == pd.core.frame.DataFrame:
            return get_dp_from_coord_df(self.data, data_name, long, lat, return_coords)
        elif type(self.data) == xr.core.dataset.Dataset:
            return get_dp_from_coord_xr(self.data, data_name, long, lat, return_coords)
    
    def get_value(self, bounds, agg_type=None, skipna=True):
        '''
        Retrieve aggregated value from within bounds
        
        Args:
            aggregation_type (str): Method of aggregation of datapoints within
                bounds. Can be upper or lower case. 
                Accepts 'MIN', 'MAX', 'MEAN', 'MEDIAN', 'STD', 'COUNT'
            bounds (Boundary): Boundary object with limits of lat/long
            skipna (bool): Defines whether to propogate NaN's or not
                Default = True (ignore's NaN's)

        Returns:
            dict: 
                {variable (str): aggregated_value (float)}
                Aggregated value within bounds following aggregation_type
                
        Raises:
            ValueError: aggregation type not in list of available methods
        '''
        def get_value_from_df(dps, bounds, agg_type, skipna):
            '''
            Aggregates a value from a pd.Series.
            
            Args:
                dps (pd.Series): Datapoints within boundary
                bounds (Boundary): 
                    Boundary dps was trimmed to. Not used for any calculations,
                    just the logging.debug message.
                agg_type (str):
                    Method of aggregation for the value, 
                    e.g. agg_type = 'MIN' => min(dps) returned 
                skipna (bool): 
                    Flag for whether NaN's should be included in aggregation. 
            
            Returns:
                np.float64: Aggregated value
            '''
            # Skip NaN's if desired
            if skipna:  dps = dps.dropna()

            logging.debug(f"\t{len(dps)} datapoints found for attribute '{self.data_name}' within bounds '{bounds}'")
            # If want the number of datapoints
            if agg_type =='COUNT':
                return len(dps)
            # If no data
            elif len(dps) == 0:
                return np.nan
            elif np.isnan(dps).all():
                return np.nan
            # Return float of aggregated value
            elif agg_type == 'MIN':
                return dps.min(skipna=skipna)
            elif agg_type == 'MAX':
                return dps.max(skipna=skipna)
            elif agg_type == 'MEAN':
                return dps.mean(skipna=skipna)
            elif agg_type == 'MEDIAN':
                return dps.median(skipna=skipna)
            elif agg_type == 'STD':
                return dps.std(skipna=skipna)
            # If aggregation_type not available
            else:
                raise ValueError(f'Unknown aggregation type {agg_type}')

        
        def get_value_from_xr(dps, bounds, agg_type, skipna):
            '''
            Aggregates a value from a xr.DataArray.
            
            Args:
                dps (xr.DataArray): Datapoints within boundary
                bounds (Boundary): 
                    Boundary dps was trimmed to. Not used for any calculations,
                    just the logging.debug message.
                agg_type (str):
                    Method of aggregation for the value, 
                    e.g. agg_type = 'MIN' => min(dps) returned 
                skipna (bool): 
                    Flag for whether NaN's should be included in aggregation. 
            
            Returns:
                np.float64: Aggregated value
            '''
            # Extract values to be worked on by numpy functions
            dps = dps.values
            logging.debug(f"\t{len(dps)} datapoints found for attribute '{self.data_name}' within bounds '{bounds}'")
            # If want the number of datapoints
            if agg_type =='COUNT':
                return dps.size
            # If no data
            elif dps.size == 0:
                return np.nan
            # If all NaN, avoid calculations
            elif np.isnan(dps).all():
                return np.nan
            # Return float of aggregated value
            elif agg_type == 'MIN':
                if skipna:  return np.nanmin(dps)
                else:       return np.min(dps)
            elif agg_type == 'MAX':
                if skipna:  return np.nanmax(dps)
                else:       return np.max(dps)
            elif agg_type == 'MEAN':
                if skipna:  return np.nanmean(dps)
                else:       return np.mean(dps)
            elif agg_type == 'MEDIAN':
                if skipna:  return np.nanmedian(dps)
                else:       return np.median(dps)
            elif agg_type == 'STD':
                if skipna:  return np.nanstd(dps)
                else:       return np.std(dps)
            elif agg_type == 'RMSE':
                if skipna:  return np.sqrt(np.nanmean((dps-np.nanmean(dps))**2))
                else:       return np.sqrt(np.mean((dps-np.mean(dps))**2))
            # If aggregation_type not available
            else:
                raise ValueError(f'Unknown aggregation type {agg_type}')

        # Set to params if no specific aggregate type specified
        if agg_type is None:
            agg_type = self.aggregate_type
            
        # Limit data series to just the data, excluding coords/index
        dps = self.trim_datapoints(bounds)[self.data_name]

        if type(self.data) == pd.core.frame.DataFrame:
            value = get_value_from_df(dps, bounds, agg_type, skipna)
        elif type(self.data) == xr.core.dataset.Dataset:
            value = get_value_from_xr(dps, bounds, agg_type, skipna)
            
        # Cast to regular float before returning so can be saved in JSON later
        return {self.data_name: float(value)}

    def get_hom_condition(self, bounds, splitting_conds):
        '''
        Retrieves homogeneity condition of data within
        boundary.
         
        Args: 
            bounds (Boundary): Boundary object with limits of datarange to analyse
            splitting_conds (dict): Containing the following keys: \n
                'threshold':  
                    `(float)` The threshold at which data points of 
                    type 'value' within this CellBox are checked to be either 
                    above or below
                'upper_bound': 
                    `(float)` The lowerbound of acceptable percentage 
                    of data_points of type value within this boundary that are 
                    above 'threshold'
                'lower_bound': 
                    `(float)` The upperbound of acceptable percentage 
                    of data_points of type value within this boundary that are 
                    above 'threshold'
                'split_lock':
                    `(bool)` If true, a cellbox will not be split by other 
                    splitting conditions if it is deemed homogeneous. default = False.

        Returns:
            str:
                The homogeniety condtion returned is of the form: \n
                'CLR' = the proportion of data points within this cellbox over a 
                given threshold is lower than the lowerbound \n
                'HOM' = the proportion of data points within this cellbox over a
                given threshold is higher than the upperbound \n
                'MIN' = the cellbox contains less than a minimum number of 
                data points \n
                'HET' = the proportion of data points within this cellbox over a
                given threshold if between the upper and lower bound
                
        '''
        def get_hom_condition_from_df(dps, splitting_conds):
            '''
            Determined homogeneity condition from pd.Series. 
            
            Args:
                dps (pd.Series):
                    Datapoints to determine homogeneity condition from
                splitting_conds (dict):
                    Key/Value pairs for threshold, upper_bound and lower_bound. 
                    More details in get_hom_condition() docstring
            
            Returns:
                str:
                    Homogeneity condition of dataset, as described in 
                    get_hom_condition() docstring
            '''
            # If not enough datapoints
            if len(dps) < self.min_dp: hom_type = "CLR"
            # Otherwise, extract the homogeneity condition
            else:
                # Determine fraction of datapoints over threshold value
                num_over_threshold = dps[dps > splitting_conds['threshold']]
                frac_over_threshold = num_over_threshold.shape[0]/dps.shape[0]

                # Return homogeneity condition
                if   frac_over_threshold <= splitting_conds['lower_bound']: hom_type = "CLR"
                elif frac_over_threshold >= splitting_conds['upper_bound']:
                    if splitting_conds['split_lock'] == True: 
                        hom_type = "HOM"
                    else: 
                        hom_type = "CLR"
                else: hom_type = "HET"

            logging.debug(f"\thom_condition for attribute: '{self.data_name}' in bounds:'{bounds}' returned '{hom_type}'")
            return hom_type
        
        def get_hom_condition_from_xr(dps, splitting_conds):
            '''
            Determined homogeneity condition from xr.DataArray. 
            
            Args:
                dps (xr.DataArray):
                    Datapoints to determine homogeneity condition from
                splitting_conds (dict):
                    Key/Value pairs for threshold, upper_bound and lower_bound. 
                    More details in get_hom_condition() docstring
            
            Returns:
                str:
                    Homogeneity condition of dataset, as described in 
                    get_hom_condition() docstring
            '''
            if dps.size < self.min_dp: 
                hom_type = "CLR"
                logging.debug(f"\t{dps.size} datapoints found for attribute '{self.data_name}' within bounds '{bounds}'")
            else:
                # Determine fraction of datapoints over threshold value
                num_over_threshold = np.count_nonzero(dps > splitting_conds['threshold'])
                frac_over_threshold = num_over_threshold/dps.size
                       
                # Return homogeneity condition
                if   frac_over_threshold <= splitting_conds['lower_bound']: hom_type = "CLR"
                elif frac_over_threshold >= splitting_conds['upper_bound']: 
                    if splitting_conds['split_lock'] == True:
                        hom_type = "HOM"
                        logging.debug(f"\tSplitting locked by attribute: '{self.data_name}' in bounds:'{bounds}'")
                    else: hom_type = "CLR"
                else: hom_type = "HET"
                
            logging.debug(f"\thom_condition for attribute: '{self.data_name}' in bounds:'{bounds}' returned '{hom_type}'")
            
            return hom_type
        
        # Set default values for splitting_conds if not provided
        if 'split_lock' not in splitting_conds:
            splitting_conds['split_lock'] = False

        dps = self.trim_datapoints(bounds)[self.data_name]
        # Retrieve datapoints to analyse
        if type(dps) == pd.core.series.Series:
            return get_hom_condition_from_df(dps, splitting_conds)
        elif type(dps) == xr.core.dataarray.DataArray:
            return get_hom_condition_from_xr(dps, splitting_conds)
        else:
            raise TypeError(f'Unknown type {type(dps)}')

        
    def reproject(self, in_proj='EPSG:4326', out_proj='EPSG:4326', 
                        x_col='lat', y_col='long'):
        '''
        Reprojects data using pyProj.Transformer
        self.data can be pd.DataFrame or xr.Dataset
        
        Args:
            in_proj (str): 
                Projection that the imported dataset is in
                Must be allowed by PyProj.CRS (Coordinate Reference System)
            out_proj (str): 
                Projection required for final data output
                Must be allowed by PyProj.CRS (Coordinate Reference System)
                Shouldn't change from default value (EPSG:4326)
            x_col (str): Name of coordinate column 1
            y_col (str): Name of coordinate column 2
                x_col and y_col will be cast into lat and long by the 
                reprojection 
            
        Returns:
            pd.DataFrame: 
                Reprojected data with 'lat', 'long' columns 
                replacing 'x_col' and 'y_col'
        '''
        def reproject_df(data, in_proj, out_proj, x_col, y_col):
            '''
            Reprojects a pandas dataframe
            
            Args:
                data (pd.DataFrame):
                    Data to reproject, with coordinates x_col, y_col
                in_proj (str): 
                    Projection the original dataset is in, as a string 
                    understandable by PyProj (e.g. 'EPSG:3031')
                out_proj (str):
                    Projection desired as output format. Should be always be
                    Mercator, but doesn't have to be if you desire
                    (e.g. 'EPSG:4326)
                x_col (str): 
                    Coordinate that original dataset includes that will be 
                    projected. Will be replaced with longitude values
                y_col (str):
                    Coordinate that original dataset includes that will be 
                    projected. Will be replaces with latitude values
                
            Returns:
                pd.DataFrame:
                    Reprojected dataset, with columns 'lat', 'long', 
                    ('time' if in original dataset), and data_name
            '''
            # Do the reprojection
            x, y = Transformer\
                    .from_crs(CRS(in_proj), CRS(out_proj), always_xy=True)\
                    .transform(data[x_col].to_numpy(), data[y_col].to_numpy())
            # Replace columns with reprojected columns called 'lat'/'long'
            if x_col != 'long':  data = data.drop(x_col, axis=1)
            if y_col != 'lat': data = data.drop(y_col, axis=1)
            data['lat']  = y
            data['long'] = x
            
            return data
            
        def reproject_xr(data, in_proj, out_proj, x_col, y_col, fast=False):
            '''
            Reprojects a xr.Dataset
            
            Args:
                data (xr.Dataset):
                    Data to reproject, with coordinates x_col, y_col
                in_proj (str): 
                    Projection the original dataset is in, as a string 
                    understandable by rioxarray (e.g. 'EPSG:3031')
                out_proj (str):
                    Projection desired as output format. Should be always be
                    Mercator, but doesn't have to be if you desire
                    (e.g. 'EPSG:4326)
                x_col (str): 
                    Coordinate that original dataset includes that will be 
                    projected. Will be replaced with longitude values
                y_col (str):
                    Coordinate that original dataset includes that will be 
                    projected. Will be replaces with latitude values
                
            Returns:
                pd.DataFrame:
                    Reprojected dataset, with columns 'lat', 'long', 
                    ('time' if in original dataset), and data_name
            '''
            if fast:
            # If want fast results (uses interpolation)
                max_size = sum(data.sizes.values())
                # Set data CRS
                data = data.rio.write_crs(in_proj)
                # Reproject
                data = data.rio.reproject(out_proj, resampling=Resampling.bilinear,
                                                    shape=((max_size,max_size)), 
                                                    nodata=np.nan)
                # Rename coordinates
                data = data.rename({x_col: 'long', y_col: 'lat'})
                # Reorder coords in case they are wrong
                data = data.sortby('lat', ascending=True)
                data = data.sortby('long', ascending=True)
                return data
            # If want accurate results
            else:
                df = data.to_dataframe().reset_index().dropna()
                return reproject_df(df, in_proj, out_proj, x_col, y_col)

        # If no reprojection to do
        if in_proj == out_proj:
            logging.debug("\tself.reproject() called but don't need to")
            return self.data
        else:
            logging.info(f"\tReprojecting data from {in_proj} to {out_proj}")
        # Choose appropriate method of reprojection based on data type
        if type(self.data) == pd.core.frame.DataFrame:
            return reproject_df(self.data, in_proj, out_proj, x_col, y_col)
        elif type(self.data) == xr.core.dataset.Dataset:
            return reproject_xr(self.data, in_proj, out_proj, x_col, y_col, 
                                fast=self.fast_reprojection)
    
    def downsample(self, agg_type=None):
        '''
        Downsamples imported data to be more easily manipulated. Data size 
        should be reduced by a factor of m*n, where (m,n) are the 
        downsample_factors defined in the params.        
        self.data can be pd.DataFrame or xr.Dataset
        
        Args:
            agg_type (str): 
                Method of aggregation to bin data by to downsample. Default is
                same method used for homogeneity condition.            

        Returns:
            xr.Dataset or pd.DataFrame: 
                Downsampled data
        '''
        def downsample_xr(data, ds, agg_type):
            '''
            Downsample xarray dataset according to aggregation type
            
            Args:
                data (xr.Dataset):
                    Dataset containing data to be downsampled. Must have 
                    coordinates 'lat' and 'long'
                ds (int, int):
                    Downsampling factors. 
                    ds[0] is longitude
                    ds[1] is latitude
                agg_type (str):
                    Aggregation method to use for binning. Default is same as 
                    set in config, passed in by parent
            
            Returns:
                xr.Dataset:
                    Downsampled data
            '''
            if agg_type == 'MIN':
                # Returns min of bin
                data = data.coarsen(lat=ds[1],boundary='pad').min()
                data = data.coarsen(long=ds[0],boundary='pad').min()
            elif agg_type == 'MAX':
                # Returns max of bin
                data = data.coarsen(lat=ds[1],boundary='pad').max()
                data = data.coarsen(long=ds[0],boundary='pad').max()
            elif agg_type == 'MEAN':
                # Returns mean of bin
                data = data.coarsen(lat=ds[1],boundary='pad').mean()
                data = data.coarsen(long=ds[0],boundary='pad').mean()
            elif agg_type == 'MEDIAN':
                # Returns median of bin
                data = data.coarsen(lat=ds[1],boundary='pad').median()
                data = data.coarsen(long=ds[0],boundary='pad').median()
            elif agg_type == 'STD':
                # Returns std_dev of range
                data = data.coarsen(lat=ds[1],boundary='pad').std()
                data = data.coarsen(long=ds[0],boundary='pad').std()
            elif agg_type =='COUNT': 
                # Returns every first element in bin
                data = data.thin(lat=ds[1])
                data = data.thin(long=ds[0])
            return data
    
        def downsample_df(data, ds, agg_type):
            '''
            Downsample pandas dataframe
            Not implemented as it just adds to processing time, 
            defeating the purpose
            '''
            logging.warning(
                '\tDownsampling called on pd.DataFrame! Downsampling a df' \
                'too computationally expensive, returning original df'
                )
            return data

        # Set to params if no specific aggregate type specified
        if agg_type is None:
            agg_type = self.aggregate_type
        
        # If no downsampling
        if self.downsample_factors == (1,1) or \
           self.downsample_factors == [1,1]:
            logging.debug("\tself.downsample() called but don't have to")
            return self.data
        else:
            logging.info(f"\tDownsampling data by {self.downsample_factors}")
        # Otherwise, downsample appropriately
        if type(self.data) == pd.core.frame.DataFrame:
            return downsample_df(self.data, self.downsample_factors, agg_type)
        elif type(self.data) == xr.core.dataset.Dataset:
            return downsample_xr(self.data, self.downsample_factors, agg_type)
        
    def get_data_col_name(self):
        '''
        Retrieve name of data column (for pd.DataFrame), or variable 
        (for xr.Dataset). Used for when data_name not defined in params.

        Returns:
            str: 
                Name of data column
            
        Raises:
            ValueError: 
                If multiple possible data columns found, can't retrieve data 
                name
        '''
        def get_data_name_from_df(data):
            '''
            Filters out standard columns to extract only data column's name 
            from pd.DataFrame
            
            Args:
                data (pd.DataFrame):
                    DataFrame with only 'lat', 'long', 'time' and data_name
            
            Returns:
                str: Data name extracted from column head
            '''
            # Store name of data column for future reference
            columns = data.columns
            # Filter out lat, long, time columns leaves us with data column name
            filtered_cols = filter(lambda col: \
                                    col not in ['lat','long','time'], columns)
            name = list(filtered_cols)
            if len(name) != 1:
                raise ValueError(
                f'More than 1 data column detected, cannot retrieve data \
                    name! Found columns: {",".join(name)}'
                                 )
            return name[0]
        
        def get_data_name_from_xr(data):
            '''
            Extracts variable name directly from xr.Dataset metadata
            
            Args:
                data (xr.Dataset):
                    Dataset with only one data variable
            
            Returns:
                str: Data name extracted from xarray metadata
            '''
            # Extract data variables from xr.Dataset
            name = list(data.keys())
            # Ensure there's only 1 data column to read name from
            if len(name) != 1:
                raise ValueError(
                f'More than 1 data column detected, cannot retrieve data \
                    name! Found columns: {",".join(name)}'
                                 )
            return name[0]
        
        logging.debug(f"\tRetrieving data name from {type(self.data)}")
        # Choose method of extraction based on data type
        if type(self.data) == pd.core.frame.DataFrame:
            return get_data_name_from_df(self.data)
        elif type(self.data) == xr.core.dataset.Dataset:
            return get_data_name_from_xr(self.data)

    def set_data_col_name(self, new_name):
        '''
        Sets name of data column/data variable
        
        Args:
            name (str): Name to replace currently stored name with

        Returns:
            xr.Dataset or pd.DataFrame: 
                Data with variable name changed
        '''
        def set_name_df(data, old_name, new_name):
            '''
            Renames data column in pandas dataframe
            
            Args:
                data (pd.DataFrame):
                    DataFrame that has old_name as one of column names
                old_name (str):
                    Old column name to replace
                new_name (str):
                    New column name after replacing            

            Returns:
                pd.DataFrame:
                    DataFrame with old_name replaced by new_name
            '''
            # Rename data column to new name
            return data.rename(columns={old_name: new_name})
        def set_name_xr(data, old_name, new_name):
            '''
            Renames data variable in xarray dataset
            
            Args:
                data (xr.Dataset):
                    Dataset that has old_name as data variable name
                old_name (str):
                    Old data_variable name to replace
                new_name (str):
                    New data_variable name after replacing            

            Returns:
                xr.Dataset:
                    Dataset with old_name replaced by new_name
            '''
            # Rename data variable to new name
            return data.rename({old_name: new_name})
        
        old_name = self.get_data_col_name()
        if old_name != new_name:
            logging.info(f"\tChanging data name from {old_name} to {new_name}")
        # Change data name depending on data type
        if type(self.data) == pd.core.frame.DataFrame:
            return set_name_df(self.data, old_name, new_name)
        elif type(self.data) == xr.core.dataset.Dataset:
            return set_name_xr(self.data, old_name, new_name)



from meshiphi.dataloaders.scalar.amsr import AMSRDataLoader
from meshiphi.dataloaders.scalar.bsose_depth import BSOSEDepthDataLoader
from meshiphi.dataloaders.scalar.bsose_sea_ice import BSOSESeaIceDataLoader
from meshiphi.dataloaders.scalar.baltic_sea_ice import BalticSeaIceDataLoader
from meshiphi.dataloaders.scalar.gebco import GEBCODataLoader
from meshiphi.dataloaders.scalar.icenet import IceNetDataLoader
from meshiphi.dataloaders.scalar.modis import MODISDataLoader
from meshiphi.dataloaders.scalar.scalar_csv import ScalarCSVDataLoader
from meshiphi.dataloaders.scalar.scalar_grf import ScalarGRFDataLoader
from meshiphi.dataloaders.scalar.shape import ShapeDataLoader
from meshiphi.dataloaders.scalar.era5_sig_wave_height import ERA5SigWaveHeightDataLoader
from meshiphi.dataloaders.scalar.era5_max_wave_height import ERA5MaxWaveHeightDataLoader
from meshiphi.dataloaders.scalar.era5_wave_period import ERA5WavePeriodDataLoader
from meshiphi.dataloaders.scalar.era5_mean_wave_direction import ERA5MeanWaveDirDataLoader
from meshiphi.dataloaders.scalar.era5_wind_mag import ERA5WindMagDataLoader
from meshiphi.dataloaders.scalar.era5_wind_dir import ERA5WindDirDataLoader

from meshiphi.dataloaders.vector.baltic_current import BalticCurrentDataLoader
from meshiphi.dataloaders.vector.era5_wind import ERA5WindDataLoader
from meshiphi.dataloaders.vector.north_sea_current import NorthSeaCurrentDataLoader
from meshiphi.dataloaders.vector.oras5_current import ORAS5CurrentDataLoader
from meshiphi.dataloaders.vector.sose import SOSEDataLoader
from meshiphi.dataloaders.vector.vector_csv import VectorCSVDataLoader
from meshiphi.dataloaders.vector.vector_grf import VectorGRFDataLoader
from meshiphi.dataloaders.vector.duacs_current import DuacsCurrentDataLoader
from meshiphi.dataloaders.vector.era5_wave_direction_vector import ERA5WaveDirectionLoader

from meshiphi.dataloaders.lut.density import DensityDataLoader
from meshiphi.dataloaders.lut.thickness import ThicknessDataLoader
from meshiphi.dataloaders.lut.scotland_ncmpa import ScotlandNCMPA
from meshiphi.dataloaders.lut.lut_csv import LutCSV
from meshiphi.dataloaders.lut.lut_geojson import LutGeoJSON
from meshiphi.dataloaders.lut.lut_shapefile import LutShapefile

from glob import glob
import os


class DataLoaderFactory:
    """
    Produces initialised DataLoader objects that can be used by the mesh to
    quickly retrieve values within a boundary.
    """
    @staticmethod
    def get_dataloader(name, bounds, params, min_dp=5):
        """
        Creates appropriate dataloader object based on name

        Args:
            name (str):
                Name of data source/type. Must be one of following -
                'scalar_csv', 'scalar_grf', 'binary_grf', 'amsr', 'bsose_sic',
                'bsose_depth', 'baltic_sic', 'gebco', 'icenet', 'modis',
                'thickness', 'density', 'circle', 'square', 'gradient',
                'checkerboard', 'vector_csv', 'vector_grf', 'baltic_currents',
                'era5_wind', 'northsea_currents', 'oras5_currents', 'sose',
                'duacs_currents', 'era5_wave_height', 'era5_wave_direction'
            bounds (Boundary):
                Boundary object with initial mesh space&time limits
            params (dict):
                Dictionary of parameters required by each dataloader
            min_dp (int):
                Minimum datapoints required to get homogeneity condition

        Returns:
            (Scalar/Vector/LUT DataLoader):
                DataLoader object of correct type, with required params set
        """
        # Cast name to lowercase to make case insensitive
        name = name.lower()
        # Translate 'file' or 'folder' into 'files' key
        params = DataLoaderFactory.translate_file_input(params)
        
        # Add loader name to params
        params['dataloader_name'] = name
        params['min_dp'] = min_dp
        
        dataloader_requirements = {
            # Scalar
            'scalar_csv':   (ScalarCSVDataLoader, ['files']),
            'scalar_grf':   (ScalarGRFDataLoader, ['binary']),
            'binary_grf':   (ScalarGRFDataLoader, ['binary']),
            'amsr':         (AMSRDataLoader, ['files', 'hemisphere']),
            'bsose_sic':    (BSOSESeaIceDataLoader, ['files']),
            'bsose_depth':  (BSOSEDepthDataLoader, ['files']),
            'baltic_sic':   (BalticSeaIceDataLoader, ['files']),
            'gebco':        (GEBCODataLoader, ['files']),
            'icenet':       (IceNetDataLoader, ['files']),
            'modis':        (MODISDataLoader, ['files']),
            'era5_sig_wave_height': (ERA5SigWaveHeightDataLoader, ['files']),
            'era5_max_wave_height': (ERA5MaxWaveHeightDataLoader, ['files']),
            'era5_wave_dir': (ERA5MeanWaveDirDataLoader, ['files']),
            'era5_wave_period': (ERA5WavePeriodDataLoader, ['files']),
            'era5_wind_mag': (ERA5WindMagDataLoader, ['files']),
            'era5_wind_dir': (ERA5WindDirDataLoader, ['files']),
            # Scalar - Abstract shapes
            'circle':       (ShapeDataLoader, []),
            'square':       (ShapeDataLoader, []),
            'gradient':     (ShapeDataLoader, []),
            'checkerboard': (ShapeDataLoader, []),
            # Vector
            'vector_csv':           (VectorCSVDataLoader, ['files']),
            'vector_grf':           (VectorGRFDataLoader, []),
            'baltic_currents':      (BalticCurrentDataLoader, ['files']),
            'era5_wind':            (ERA5WindDataLoader, ['files']),
            'era5_wave_direction':  (ERA5WaveDirectionLoader, ['files']),
            'northsea_currents':    (NorthSeaCurrentDataLoader, ['files']),
            'duacs_currents':       (DuacsCurrentDataLoader, ['files']),
            'oras5_currents':       (ORAS5CurrentDataLoader, ['files']),
            'sose':                 (SOSEDataLoader, ['files']),
            # LUT
            'thickness':        (ThicknessDataLoader, []),
            'density':          (DensityDataLoader, []),
            'scotland_ncmpa':   (ScotlandNCMPA, ['files']),
            'lut_csv':          (LutCSV, ['files']),
            'lut_geojson':      (LutGeoJSON, ['files', 'value']),
            'lut_shapefile':    (LutShapefile, ['files', 'value'])
        }
        # If name is recognised as a dataloader
        if name in dataloader_requirements:
            # Set data loader and params required for it to work
            data_loader = dataloader_requirements[name][0]
            required_params = dataloader_requirements[name][1]
        else: 
            raise ValueError(f'{name} not in known list of DataLoaders')

        # Assert dataloader will get all required params to work
        assert all(key in params for key in required_params), \
            f'Dataloader {name} is missing some parameters! Requires {required_params}. Has {list(params.keys())}'

        # Create instance of dataloader
        return data_loader(bounds, params)

    @staticmethod
    def translate_file_input(params):
        """
        Allows flexible file specification in params. Translates 'file' or
        'folder' into 'files'

        Args:
            params (dict):
                Dictionary of parameters written in config
        """
        if 'file' in params:
            params['files'] = [params['file']]
            del params['file']
        elif 'folder' in params:
            folder = os.path.join(params['folder'], '')  # Adds trailing slash if non-existent
            params['files'] = sorted(glob(folder+'*'))
            del params['folder']
        return params



from meshiphi.dataloaders.lut.abstract_lut import LutDataLoader

import logging

import pandas as pd
from shapely import wkt

class DensityDataLoader(LutDataLoader):
    def import_data(self, bounds):
        '''
        Creates a simulated dataset of sea ice density based on 
        scientific literature.
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            density_df (pd.DataFrame): 
                Sea Ice Density dataset within limits of bounds. 
                DataFrame has coordinates 'date', 'shape', 
                and variable 'density'
        '''
        
        # Look up table parameters hardcoded
        northern_hemisphere = wkt.loads('POLYGON((-180 0, -180 90, 180 90, 180 0, -180 0))')
        southern_hemisphere = wkt.loads('POLYGON((-180 -90, -180 0, 180 0, 180 -90, -180 -90))')
        
        northern_seasons = {
            1: 'wi',  2: 'wi', 12: 'wi',
            3: 'sp',  4: 'sp',  5: 'sp', 
            6: 'su',  7: 'su',  8: 'su', 
            9: 'au', 10: 'au', 11: 'au',
            }
        southern_seasons = {
            1: 'su',  2: 'su', 12: 'su',
            3: 'au',  4: 'au',  5: 'au', 
            6: 'wi',  7: 'wi',  8: 'wi', 
            9: 'sp', 10: 'sp', 11: 'sp',
            }
        
        densities = {
            'su': 875.0, 
            'sp': 900.0, 
            'au': 900.0, 
            'wi': 920.0
            }
        
        # Create shape to intersect regions with
        bounds_polygon = bounds.to_polygon()
        
        # For every date in range, create a new shape and value pair
        dates = pd.date_range(start=bounds.get_time_min(), end=bounds.get_time_max())
        
        density_df = pd.concat(
                        [pd.DataFrame(
                            {'time': dates,
                            'geometry': northern_hemisphere & bounds_polygon,    # Intersect shapes
                            'density': [densities[northern_seasons[month]] for month in dates.month]}),
                        pd.DataFrame(
                            {'time': dates,
                            'geometry': southern_hemisphere & bounds_polygon,    # Intersect shapes
                            'density': [densities[southern_seasons[month]] for month in dates.month]})
                        ]
        )
        density_df = density_df.set_index('time').sort_index()
        return density_df


import math
from scipy.stats import qmc
import numpy as np

class Sampler:
    """
    a class that generates samples based on Sobol sequences, which fills the space in a highly uniform way and guarantees a better coverage of the sampling space. 
    Attributes:
        dimensions (int): an integer representing the dimensions of each sample 
        number_of_samples (int): an integer representing the number of the generated samples


    """
    def __init__(self, d , n ):
      
        self.dimensions = d
        self.number_of_samples = n


    def generate_samples ( self, ranges):
        """

          generates samples within the provided ranges array, the length of the ranges should equal to self.dimensions

            Args:
              ranges (float[]): an array that contains the range that each sample dimension should fall within
            Returns:
                distance (float): the RMSE between the actaul data value and the mesh's agg_value.

        """
        if len(ranges) != self.dimensions:
            raise ValueError("ranges length should be equal to the sampler dimension") 

        sampler =  qmc.Sobol(d=self.dimensions)
        samples = sampler.random(n=self.number_of_samples)
        mapped_samples = []
        # map samples to ranges
        for sample in samples:
            for i in range (len(ranges)):
                mapped_samples. append (ranges [i] [0] + sample[i]* (ranges[i] [1] - ranges [i][0]))
        return np.array(mapped_samples).reshape ( len (samples), len(ranges))
        
        
       



from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader

import xarray as xr

class BSOSEDepthDataLoader(ScalarDataLoader):
    def import_data(self, bounds):
        """
        Reads in data from a BSOSE Depth NetCDF file.
        Renames coordinates to 'lat' and 'long', and renames variable to
        'elevation'

        Args:
            bounds (Boundary): Initial boundary to limit the dataset to

        Returns:
            xr.Dataset:
                BSOSE Depth dataset within limits of bounds.
                Dataset has coordinates 'lat', 'long', and variable 'elevation'
        """
        # Open Dataset
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files)
        # Change column names
        data = data.rename({'Depth': 'elevation',
                            'YC': 'lat',
                            'XC': 'long'})
        # Change domain of dataset from [0:360) to [-180:180)
        data = data.assign_coords(long=((data.long + 180) % 360) - 180)
        # Sort the 'long' axis so that sel() will work
        data = data.sortby('long')
        # Convert elevation from coord to variable
        data = data.reset_coords('elevation')
        # Convert depth to elevation by inverting sign
        data['elevation'] = -1 * data['elevation']
        # Limit to just elevation data
        data = data['elevation'].to_dataset()
        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)
        
        return data



class Direction:
    north_east=1
    east=2
    south_east=3
    south=4
    south_west=-1
    west=-2
    north_west=-3
    north= -4

    def __init__(self):
        self.north_east = 1
        self.east = 2
        self.south_east = 3
        self.south = 4
        self.south_west = -1
        self.west = -2
        self.north_west = -3
        self.north = -4



from meshiphi.dataloaders.scalar.abstract_scalar import ScalarDataLoader

import pandas as pd

class ScalarCSVDataLoader(ScalarDataLoader):
    def import_data(self, bounds):
        '''
        Reads in data from a CSV file. 
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            pd.DataFrame: 
                Scalar dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', potentially 'time',
                and variable defined by column heading in csv file
        '''
        # Read in data
        df_list = []
        # NOTE: All csv files must have same columns for this to work
        for file in self.files:
            df_list += [pd.read_csv(file)]
        data = pd.concat(df_list)
        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)
        
        return data



from abc import ABCMeta, abstractmethod
from meshiphi.mesh_generation.boundary import Boundary

class DataLoaderInterface(metaclass=ABCMeta):
    '''
    The interface through which the mesh operates on the dataloaders.
    Dataloaders store all data internally, and will retrieve
    aggregated values and homogeneity conditions when called.
    '''
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_value')          and
                callable(subclass.get_value)            and
                hasattr(subclass, 'get_hom_condition')  and
                callable(subclass.get_hom_condition)   or
                NotImplemented)
    
    @abstractmethod  
    def get_value(self, bounds: Boundary, agg_type: str):
        ''' Reads values within boundary and returns aggregated value as 
        a dictionary in the form {data_name (str): value (np.float64)}'''
        raise NotImplementedError

    @abstractmethod  
    def get_hom_condition(self, bounds: Boundary, splitting_conds: dict):
        ''' Reads values within boundary to determine if data is
        homogeneous or heterogeneous. Return as str'''
        raise NotImplementedError


import json
import logging
import geopandas as gpd
import pandas as pd
from shapely import wkt
import numpy as np
import subprocess
import sys
import os
import tempfile

from meshiphi.mesh_generation.jgrid_aggregated_cellbox import JGridAggregatedCellBox
from meshiphi.mesh_generation.boundary import Boundary
from meshiphi.mesh_generation.aggregated_cellbox import AggregatedCellBox
from meshiphi.mesh_generation.neighbour_graph import NeighbourGraph


from meshiphi.mesh_validation.sampler import Sampler
import collections.abc
import math
from pathlib import Path

class EnvironmentMesh:
    """
        a class that defines the environmental mesh structure and contains each cellbox aggregate information

        Attributes:
            bounds (Boundary): the boundaries of this mesh 
            agg_cellboxes (AggregatedCellBox[]): a list of aggregated cellboxes
            neighbour_graph(NeighbourGraph): an object contains each cellbox neighbours information 
            config (dict): conatins the initial config used to build this mesh\n
    """
    @classmethod
    def load_from_json(cls, mesh_json):
        """
            Constructs an Env.Mesh from a given env-mesh json file to be used by other modules (ex.Vessel Performance Modeller).

            Args:
                mesh_json (dict): a dictionary loaded from an Env-mesh json file of the following format - \n
                    \n
                    {\n
                        "mesh_info": {\n
                            "region": {\n
                                "lat_min": (real),\n
                                "lat_max": (real),\n
                                "long_min": (real),\n
                                "long_max": (real),\n
                                "start_time": (string) 'YYYY-MM-DD',\n
                                "end_time": (string) 'YYYY-MM-DD',\n
                                "cell_width": (real),\n
                                "cell_height" (real),\n
                                "split_depth" (int)\n
                            },\n
                            "data_sources": [\n
                                {\n
                                    "loader": (string)\n
                                    "params" (dict)\n
                                },\n
                                ...,\n
                                {...}
                                ], \n
                            "splitting": {\n
                                    "split_depth": (int),\n
                                    "minimum_datapoints": (int)\n
                            }\n
                        },\n
                        "cellboxes": [\n
                            {\n

                            },\n
                            ...,\n
                            {...}\n

                        ],\n
                        "neighbour_graph": [\n
                            {\n

                            },\n
                            ...,\n
                            {...}\n
                        ]\n
                    }\n
            Returns:
                EnvironmentMesh: object that contains all the json file mesh information.\n
        """
        config = mesh_json['config']['mesh_info']
        cellboxes_json = mesh_json['cellboxes']
        agg_cellboxes = []
        bounds = Boundary.from_json(config)
        # load the agg_cellboxes
        for cellbox_json in cellboxes_json:
            agg_cellbox = AggregatedCellBox.from_json(cellbox_json)
            agg_cellboxes.append(agg_cellbox)
        neighbour_graph = NeighbourGraph.from_json(
            mesh_json['neighbour_graph'])
        obj = EnvironmentMesh(bounds, agg_cellboxes, neighbour_graph, config)
        return obj

    def __init__(self, bounds, agg_cellboxes, neighbour_graph, config):
        """
            Args:
              bounds (Boundary): the boundaries of this mesh 
              agg_cellboxes (AggregatedCellBox[]): a list of aggregated cellboxes
              neighbour_graph(NeighbourGraph): an object contains each cellbox neighbours information 
              config (dict): conatins the initial config used to build this mesh.\n
        """

        self.bounds = bounds
        self.agg_cellboxes = agg_cellboxes
        self.neighbour_graph = neighbour_graph
        self.config = config

    def to_json(self):
        """
            Returns this Mesh converted to a JSON object.\n

            Returns:
                json: a string representation of the CellGird parseable as a JSON object. The JSON object is of the form -\n

                    {\n
                        "config": the config used to initialize the Mesh,\n
                        "cellboxes": a list of CellBoxes contained within the Mesh,\n
                        "neighbour_graph": a graph representing the adjacency of CellBoxes within the Mesh.\n
                    }\n
        """
        output = dict()
        output['config'] = {'mesh_info': self.config}
        output["cellboxes"] = self.cellboxes_to_json()
        output['neighbour_graph'] = self.neighbour_graph.get_graph()


        return json.loads(json.dumps(output, indent=4))
    
    def to_png(self, params_file, path):
        """
            exports a mesh and saves to a png file
            Args:
                params_file: A format configuration files as a json object
                path: The path to save the png file
        """
        from meshiphi.mesh_plotting.mesh_plotter import MeshPlotter

        mesh_json = self.to_json()
        plotter = MeshPlotter(mesh_json)

        if (params_file != None):
            with open(params_file) as f:
                data = f.read()
                format_params = json.loads(data)

        for item in format_params['cmaps']:
            plotter.plot_cmap(item['data_name'], item['cmap'])
        for item in format_params['bools']:
            plotter.plot_bool(item['data_name'], item['colour'])
            
        plotter.save(path)


    def to_geojson(self, params_file = None):
        """
            Returns the cellboxes of this mesh converted to a geoJSON format.\n

            Returns:
                geojson: The cellboxes of this mesh in a geoJSON format

            NOTE:
                geoJSON format does not contain all the data included in the standard 
                .to_json() format. geoJSON meshes do not contain the configs used to 
                build them, or the neighbour-graph which details how each of the 
                cellboxes are connected together.\n
        """
        geojson = ""
        mesh_json = self.to_json()

        if (params_file != None):
            with open(params_file) as f:
                data = f.read()
                format_params = json.loads(data)
            data_name = format_params['data_name']
            logging.info("exporting layer : " + str(data_name))

        # Formatting mesh to geoJSON
        mesh_df = pd.DataFrame(mesh_json['cellboxes'])

        for column in mesh_df.columns:

            if column in ['id', 'geometry']:
                continue
            # remove unnecessary columns
            if (params_file != None) and column not in [str(data_name)]:
                mesh_df = mesh_df.drop(column, axis=1)
            # remove unnecessary columns
            elif column in ['cx', 'cy', 'dcx', 'dcy']:
                mesh_df = mesh_df.drop(column, axis=1)
            # convert lists to mean
            elif mesh_df[column].dtype == list:
                mesh_df[column] = [np.mean(x) for x in mesh_df[column]]
            # convert bools to ints
            elif mesh_df[column].dtype == bool:
                mesh_df[column] = mesh_df[column].astype(int)
                mesh_df[column] = mesh_df[column].replace(0, np.nan)
            
        # Remove infs and replace with nan
        mesh_df = mesh_df.replace([np.inf, -np.inf], np.nan)

        mesh_df['geometry'] = mesh_df['geometry'].apply(wkt.loads)
        mesh_gdf = gpd.GeoDataFrame(
            mesh_df, crs="EPSG:4326", geometry="geometry")
        geojson = json.loads(mesh_gdf.to_json())

        return geojson

    def to_tif(self, params_file,  path):
        """
                generates a representation of the mesh in geotif image format.\n

                Args:
                    params_file(string) (optional): a path to a file that contains a dict of the folowing export parameters (If not given, default values are used for image export). The file should be of the following format -\n
                            \n
                            {\n
                                "data_name": "elevation",\n
                                "sampling_resolution": [\n
                                    150,\n
                                    150\n
                                 ],\n
                                "projection": "3031",\n
                            }\n
                            Where data_name (string) is the name of the mesh data that will be included in the tif image (ex. SIC, elevation), if it is a vector data (e.g. fuel) then the vector mean is calculated for each pixel,
                            sampling_resolution ([int]) is a 2d array that represents the sampling resolution the geotiff will be generated at (how many pixels in the final image),
                            projection (int) is an int representing the ESPG sampling projection used to create the geotiff image  (default is 4326),
                            and colour_conf (string) contains the path to color config file, which is a text-based file containing the association between data_name values and colors. It contains 4 columns per line: the data_name value and the corresponding red, green, blue value between 0 and 255, an example format where values range from 0 to 100 is -\n
                                    \n
                                    0 240 250 160  \n
                                    30 230 220 170  \n
                                    60 220 220 220 \n
                                    100 250 250 250  \n
                    path (string): the path to save the generated tif image.\n
        """
        def generate_samples():
            """
                generates uniform lat, long samples covering the image resolution space.\n

                Returns:
                    samples([[lat,long],..]): an array of samples, each item in the array is a 2d array that contains each sample lat and long.\n

            """
            mesh_height = self.bounds.get_lat_max() - self.bounds.get_lat_min()
            mesh_width = self.bounds.get_long_max() - self.bounds.get_long_min()
            pixel_height = mesh_height / ncols
            pixel_width = mesh_width / nlines
            samples = []
            # has to move in this direction as we start rendering from the upper left pixel
            for lat in np.arange(self.bounds.get_lat_max(), self.bounds.get_lat_min(), -1*pixel_height):
                for long in np.arange(self.bounds.get_long_min(), self.bounds.get_long_max(), pixel_width):
                    pixel_lat = lat - 0.5 * pixel_height   # centeralize the pixel lat value
                    pixel_long = long + 0.5*pixel_width   # centeralize the pixel long value
                    samples = np.append(samples, pixel_lat)
                    samples = np.append(samples, pixel_long)
            # shape the samples in 2d array (each entry in the array holds sample lat and long
            samples = np.reshape(samples, (nlines * ncols, 2))
            return samples

        def get_sample_value(sample):
            """
                finds the aggregated cellbox that contains the sample lat and long and returns the value within.\n

                Args:
                    sample ([lat,long]): an array conatins the sample latitude and longtitude
                Returns:
                    the aggregated value of 'data_name'(specified in to_tif params) 

             """
            lat = sample[0]
            long = sample[1]
            value = np.nan
            for agg_cellbox in self.agg_cellboxes:
                if agg_cellbox.contains_point(lat, long):
                    # get the agg_value
                    try:
                        value = agg_cellbox.agg_data[data_name]
                    except KeyError:
                        logging.debug(f'{data_name} not found in cellbox!')
                        value = np.nan
                        
                    if isinstance(value, collections.abc.Sequence): # if it is a vector then take the mean
                        value = np.mean (value)
                        if value == float('inf') : # repalce inf with nan
                            value = np.nan
                    # break to make sure we avoid getting multiple values (for lat and long on the bounds of multiple cellboxes)
                    break
            return value

        def get_geo_transform(extent, nlines, ncols):
            """
                transforms from the image coordinate space (row, column) to the georeferenced coordinate space. \n
                Returns:    
                  GT : array consists of 6 items representing how GDAL would place the top left pixel on the generated Geotiff:\n
                  GT[0] x-coordinate of the upper-left corner of the upper-left pixel.\n
                  GT[1] w-e pixel resolution / pixel width.\n
                  GT[2] row rotation (typically zero).\n
                  GT[3] y-coordinate of the upper-left corner of the upper-left pixel.\n
                  GT[4] column rotation (typically zero).\n
                  GT[5] n-s pixel resolution / pixel height (negative value for a north-up image).\n

            """
            resx = (extent[2] - extent[0]) / ncols
            resy = (extent[3] - extent[1]) / nlines
            return [extent[0], resx, 0, extent[3], 0, -resy]

        def load_params(params_file):
            """
                  loads the parameters of the tif export and override the default values.\n

                  Args:
                      params_file (string): a path to a file containing a dict of the export params 
                  
                  Returns:
                      params (dict): a dict object that contains the loaded parameters

            """
            params = {"data_name": "SIC", "sampling_resolution": [
                100, 100], "projection": "4326"}  # the default values
            if (params_file != None):
                with open(params_file) as f:
                    data = f.read()
                    input_params = json.loads(data)
                if (input_params != None):
                    if ("projection" in input_params.keys()):
                        params["projection"] = input_params["projection"]
                    if ("data_name" in input_params.keys()):
                        params["data_name"] = input_params["data_name"]
                    if ("sampling_resolution" in input_params.keys()):
                        params["sampling_resolution"] = input_params["sampling_resolution"]
                    if ("colour_conf" in input_params.keys()):
                        params["colour_conf"] = input_params["colour_conf"]
            return params

        def transform_proj(path, params, default_proj):
            """
                  method that transforms the generated tif into another projection

                  Args:
                        path(string): the path of the generated tif
                        params (dict): a dict that contains the export parametrs
                        default_proj (stirng): a string represents the default projection (EPSG:4326).\n

            """
            if params["projection"] != str(default_proj):
                dest = osr.SpatialReference()
                dest.ImportFromEPSG(int(params["projection"]))
                # transform to target proj and save
                gdal.Warp(str(path),  str(path), dstSRS=dest.ExportToWkt())
        
        def set_colour(data, input_file, params):
            """
                  method that changes the color of the generated tif instead of using the default greyscale.
                  It defines a scale of RGB colors based on the range of data values(an example file is in unit_tests/resources/colour_conf.txt).\n

                  Args:
                        data ([float]): an array conatins the values of the 'data_name'
                        input_path(string): the path of the generated grey tif
                        params (dict): a dict that contains the export parametrs.\n

            """
            fp, color_file = tempfile.mkstemp(suffix='.txt')
            data = data[~np.isnan(data)]  # get rid of the nans before calculating range
            _max = np.nanmax(data)
            _min = np.nanmin(data)
            _range = _max-_min
            inf_color = '255 0 0' # red color for inf value
            colors = ['255 255 255', '173 216 230', '0 0 128', '0 0 139']  # default color
            with open(color_file, 'w') as f:
                for i, c in enumerate(colors[:-1]):
                    f.write(str(int(_min + (i + 1)*_range/len(colors))) +
                            ' ' + c + '\n')
                f.write(str(int(_max - _range/len(colors))) +
                        ' ' + colors[-1] + '\n')
                f.write (str(np.nan) +
                        ' ' + inf_color + '\n') # render nans in red
            os.close(fp)
            if "colour_conf" in params.keys():
                color_file = params["colour_conf"]
            cmd = "gdaldem color-relief " + input_file \
                + ' ' + color_file + ' ' + input_file
            subprocess.check_call(cmd, shell=True)
            # remove additional files created while generating the tif
            file_path = os.path.abspath(input_file)
            dir_path = os.path.split(file_path)[0]
            additional_files = ["grid_data" , "grid_data.aux.xml"]
            for file in additional_files: 
                file_path = Path(dir_path +"/"+ file)
                if os.path.isfile (file_path):
                     os.remove(file_path)

        # Only import if we need GDAL, to avoid having it as a requirement
        from osgeo import gdal, ogr, osr
        
        params = {}
        params = load_params(params_file)
        data_name = params["data_name"]
        DEFAULT_PROJ = 4326

        # Get image dimensions
        nlines = params["sampling_resolution"][0]
        ncols = params["sampling_resolution"][1]

        # define image extent based on mesh bounds
        extent = [self.bounds.get_long_min(), self.bounds.get_lat_min(
        ), self.bounds.get_long_max(), self.bounds.get_lat_max()]

        logging.info("Generating the tif image ...")
        samples = generate_samples()
        # create raster band and populate with sampled data of image_size (sampling_resolution)
        driver = gdal.GetDriverByName('GTiff')
        # reading the samples value
        data = []
        data = np.reshape(np.asarray([get_sample_value(
            sample) for sample in samples], dtype=np.float32), (nlines, ncols))
        # create a temp grid
        grid_data = driver.Create(
            'grid_data', ncols, nlines, 1, gdal.GDT_Float32)
        # setup geo-transform
        grid_data.SetGeoTransform(get_geo_transform(extent, nlines, ncols))
        # Write data
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(DEFAULT_PROJ)
        grid_data.SetProjection(srs.ExportToWkt())
        grid_data.GetRasterBand(1).WriteArray(data)
    

        # Save the file
        driver.CreateCopy(path, grid_data, 0)
        transform_proj(path, params, DEFAULT_PROJ)
        set_colour(data, path, params)
        logging.info(f'Generated GeoTIFF: {path}')

    def cellboxes_to_json(self):
        """
            returns a list of dictionaries containing information about each cellbox
            in this Mesh.
            all cellboxes will include id, geometry, cx, cy, dcx, dcy

            Returns:
                cellboxes (list<dict>): a list of CellBoxes which form the Mesh.
                    CellBoxes are of the form -

                    {
                        "id": (string) ... \n
                        "geometry": (string) POLYGON(...), \n
                        "cx": (float) ..., \n
                        "cy": (float) ..., \n
                        "dcx": (float) ..., \n
                        "dcy": (float) ..., \n
                        \n
                        "value_1": (float) ..., \n
                        ..., \n
                        "value_n": (float) ... \n
                    }
        """

        cellboxes_json = []
        for cellbox in self.agg_cellboxes:

            # Get json for CellBox
            cell = cellbox.to_json()

            cellboxes_json.append(cell)
        return cellboxes_json

    def update_cellbox(self, index, values):
        """
            method that adds values to the dict of a cellbox at certain index (to be used by the vessel perf. module to add the perf. metrics to the cellbox)

            Args:
              index (int): the index of the cellbox to be updated
              values (dict): a dict contains perf. metrics names and values
        """
        if index > -1 or index < len(self.agg_cellboxes):
            self.agg_cellboxes[index].agg_data.update(values)
        else:
            raise ValueError('Invalid cellbox index')

    def save(self, path, format="JSON", format_params=None):
        """
            Saves this object to a location in local storage in a specific format. 

            Args:
                path (String): The file location the mesh will be saved to.
                format (String) (optional): The format the mesh will be saved in.
                    If not format is given, default is JSON.
                    Supported formats are\n
                        - JSON \n
                        - GEOJSON
        """

        logging.info(f"Saving mesh in {format} format to {path}")
        if format.upper() == "TIF":
            self.to_tif(format_params, path)

        elif format.upper() == "JSON":
            with open(path, 'w') as path:
                json.dump(self.to_json(), path, indent=4)
           
        elif format.upper() == "GEOJSON":
            with open(path, 'w') as path:
                json.dump(self.to_geojson(format_params), path, indent=4)

        elif format.upper() == "PNG":
            self.to_png(format_params, path)

        else:
            logging.warning(f"Cannot save mesh in a {format} format")

        if isinstance(self.agg_cellboxes[0], JGridAggregatedCellBox):
            dump_path = path.replace(".json", ".dump")
            with open(dump_path, 'w') as dump_f:
                self.dump_mesh(dump_f)


    def dump_mesh(self, file):
        """
            creates a string representation of this Mesh which
            is then saved to a file location specified by parameter
            'file' for use in j_grid regression testing,

        """
        mesh_dump_str = ""
        for cell_box in self.agg_cellboxes:
            if isinstance(cell_box, JGridAggregatedCellBox):
                mesh_dump_str += cell_box.mesh_dump()

        file.write(mesh_dump_str)
        file.close()



import argparse
import json
import inspect
import logging

from meshiphi import __version__ as version
from meshiphi.utils import setup_logging, timed_call, convert_decimal_days
from meshiphi.mesh_generation.mesh_builder import MeshBuilder
from meshiphi.mesh_generation.environment_mesh import EnvironmentMesh


@setup_logging
def get_args(
        default_output: str,
        config_arg: bool = True,
        mesh_arg: bool = False,
        format_arg: bool = False):
    """
    Adds required command line arguments to all CLI entry points.

    Args:
        default_output (str): The default output file location.
        config_arg (bool): True if the CLI entry point requires a <config.json> file. Default is True.
        mesh_arg (bool): True if the CLI entry point requires a <mesh.json> file. Default is False.

    Returns:

    """
    ap = argparse.ArgumentParser()

    ap.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=version))

    # Optional arguments used in all CLI entry points
    ap.add_argument("-o", "--output",
                    default=default_output,
                    help="Output file")
    ap.add_argument("-v", "--verbose",
                    default=False,
                    action="store_true",
                    help="Turn on DEBUG level logging")

    if config_arg:
        ap.add_argument("config", type=argparse.FileType("r"), 
                    help="File location of a <config.json> file")

    if mesh_arg:
        ap.add_argument("mesh", type=argparse.FileType("r"),
                    help="File location of the environmental mesh")

        
    if format_arg:
        ap.add_argument("format",
                        help = "Export format to transform a mesh into. Supported \
                        formats are JSON, GEOJSON, Tif")
        ap.add_argument( "-f", "--format_conf",
                        default = None,
                        help = "File location of Export to Tif configuration parameters")



    return ap.parse_args()

@timed_call
def rebuild_mesh_cli():
    """
        CLI entry point for rebuilding the mesh based on its encoded config files.
    """

    default_output = "rebuild_mesh.output.json"
    args = get_args(default_output, mesh_arg=True, config_arg=False)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    mesh_json = json.load(args.mesh)
    config = mesh_json['config']['mesh_info']

    # rebuilding mesh...
    rebuilt_mesh = MeshBuilder(config).build_environmental_mesh()
    rebuilt_mesh_json = rebuilt_mesh.to_json()

    logging.info("Saving mesh to {}".format(args.output))
    
    json.dump(rebuilt_mesh_json, open(args.output, "w"), indent=4)



@timed_call
def create_mesh_cli():
    """
        CLI entry point for the mesh construction
    """
    
    default_output = "create_mesh.output.json"
    args = get_args(default_output)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    config = json.load(args.config)

    # Discrete Meshing
    cg = MeshBuilder(config).build_environmental_mesh()

    logging.info("Saving mesh to {}".format(args.output))
    info = cg.to_json()
    json.dump(info, open(args.output, "w"), indent=4)
    

@timed_call
def export_mesh_cli():
    """
        CLI entry point for exporting a mesh to standard formats.
        Currently supported formats are JSON, GEOJSON, TIF, PNG
    """
    # Default, used only by the MeshiPhi and PolarRoute
    args = get_args("mesh.json", 
                    config_arg = False, 
                    mesh_arg = True, 
                    format_arg = True)
        
    if args.format.upper() == "GEOJSON":
        args = get_args("mesh_geo.json", 
                    config_arg = False, 
                    mesh_arg = True, 
                    format_arg = True)
        
    elif args.format.upper() == "TIF":
        args = get_args("mesh.tif", 
                    config_arg = False, 
                    mesh_arg = True, 
                    format_arg = True)

    elif args.format.upper() == "PNG":
        args = get_args("mesh.png", 
                    config_arg = False, 
                    mesh_arg = True, 
                    format_arg = True)
    
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    mesh = json.load(args.mesh)
    env_mesh = EnvironmentMesh.load_from_json(mesh)

    logging.info(f"exporting mesh to {args.output} in format {args.format}")

    env_mesh.save(args.output, args.format , args.format_conf)
